"""Core jobs for reaction network creation and analysis."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable

import numpy as np
import ray
from jobflow import SETTINGS, Maker, job
from pymatgen.core.composition import Element
from tqdm import tqdm

from rxn_network.core import Composition
from rxn_network.costs.calculators import (
    ChempotDistanceCalculator,
    PrimaryCompetitionCalculator,
    SecondaryCompetitionCalculator,
)
from rxn_network.costs.functions import Softplus
from rxn_network.entries.utils import get_all_entries_in_chemsys, process_entries
from rxn_network.enumerators.utils import get_computed_rxn, run_enumerators
from rxn_network.jobs.schema import (
    CompetitionTaskDocument,
    EntrySetDocument,
    EnumeratorTaskDocument,
    NetworkTaskDocument,
    PathwaySolverTaskDocument,
)
from rxn_network.jobs.utils import get_added_elem_data
from rxn_network.network.network import ReactionNetwork
from rxn_network.pathways.solver import PathwaySolver
from rxn_network.reactions.basic import BasicReaction
from rxn_network.reactions.hull import InterfaceReactionHull
from rxn_network.reactions.reaction_set import ReactionSet
from rxn_network.utils.funcs import get_logger, grouper
from rxn_network.utils.ray import initialize_ray, to_iterator

if TYPE_CHECKING:
    from rxn_network.costs.base import CostFunction
    from rxn_network.entries.entry_set import GibbsEntrySet
    from rxn_network.enumerators.base import Enumerator
    from rxn_network.pathways.pathway_set import PathwaySet

logger = get_logger(__name__)


@dataclass
class GetEntrySetMaker(Maker):
    """
    Maker to create job for acquiring and processing entries to be used in reaction
    enumeration or network building. See GibbsEntrySet for more information about how
    these are constructed.

    Args:
        name: Name of the job.
        temperature: Temperature [K] for determining Gibbs Free Energy of
            formation, dGf(T).
        e_above_hull: Energy above hull (eV/atom) for thermodynamic stability threshold;
            i.e., include all entries with energies below this value.
        filter_at_temperature: Temperature (in Kelvin) at which entries are filtered for
            thermodynamic stability (e.g., room temperature). Generally, this often
            differs from the synthesis temperature.
        include_nist_data: Whether to include NIST-JANAF data in the entry set.
            Defaults to True.
        include_freed_data: Whether to include FREED data in the entry set. Defaults
            to False. WARNING: This dataset has not been thoroughly tested. Use at
            your own risk!
        formulas_to_include: An iterable of compositional formulas to ensure are
            included in the processed dataset. Sometimes, entries are filtered out that
            one would like to include, or entries don't exist for those compositions.
        calculate_e_above_hulls: Whether to calculate e_above_hull and store as an
            attribute in the data dictionary for each entry. Defaults to True.
        ignore_nist_solids: Whether to ignore NIST data for solids with high melting
            points (Tm >= 1500 ºC). Defaults to True.
        custom_entries: An optional list of user-created entries that will be added to
            the final entry set.
        MP_API_KEY: The Materials Project API key to use for querying MP using mp-api.
            If not provided, this will default to whatever is stored through
            configuration settings on the user's machine.
        entry_db_name (for internal use only): if not None, then this method will use an
            internal materials databse using the get_all_entries_in_chemsys() method. If
            you wish to use this approach, you may need to reconfigure that method.
        property_data: a list of attributes to ensure are included in each entry's data,
            if available. This currently only applies to those using a custom entry DB
            (see above).
    """

    name: str = "get_and_process_entries"
    temperature: int = 300
    e_above_hull: float = 0.0
    filter_at_temperature: int | None = None
    include_nist_data: bool = True
    include_freed_data: bool = False
    include_polymorphs: bool = False
    formulas_to_include: list[str] = field(default_factory=list)
    calculate_e_above_hulls: bool = True
    ignore_nist_solids: bool = True
    custom_entries: list = field(default_factory=list)
    MP_API_KEY: str | None = None
    entry_db_name: str | None = None
    property_data: list[str] | None = None

    @job(entries="entries", output_schema=EntrySetDocument)
    def make(self, chemsys: str):
        """
        Returns a job that acquires a GibbsEntrySet for the desired chemical system.

        NOTE: This job stores the entry set in an additional store called
        "entries". This needs to be configured through a user's jobflow.yaml file. See
        "additional_stores".

        Args:
            chemsys: The chemical system of the entry set to be acquired.
        """
        entry_db = SETTINGS.JOB_STORE.additional_stores.get(self.entry_db_name)

        if entry_db:
            logger.info(f"Using user-specified Entry DB: {self.entry_db_name}")
            property_data = self.property_data
            if property_data is None:
                property_data = ["theoretical"]
            elif "theoretical" not in property_data:
                property_data.append("theoretical")

            entries = get_all_entries_in_chemsys(
                entry_db,
                chemsys,
                compatible_only=True,
                property_data=property_data,
                use_premade_entries=False,
            )
        else:
            try:
                from mp_api.client import MPRester
            except ImportError as err:
                raise ImportError(
                    "You may need to install the Materials Project API: pip install -U"
                    " mp-api"
                ) from err

            kwargs = {}
            if self.MP_API_KEY:
                kwargs["api_key"] = self.MP_API_KEY

            elems = {Element(i) for i in chemsys.split("-")}

            if len(elems) <= 5:
                with MPRester(**kwargs) as mpr:
                    entries = mpr.get_entries_in_chemsys(
                        elements=chemsys,
                        additional_criteria={"thermo_types": ["GGA_GGA+U"]},
                    )
            else:  # this approach is faster for big systems
                other_elems = self._get_exclude_elems(elems)

                with MPRester(**kwargs) as mpr:
                    docs = mpr.summary.search(
                        exclude_elements=other_elems, all_fields=False, deprecated=False
                    )
                mpids = [d.material_id for d in docs]

                with MPRester(**kwargs) as mpr:
                    entries = mpr.get_entries(
                        mpids, additional_criteria={"thermo_types": ["GGA_GGA+U"]}
                    )

        if self.custom_entries:
            entries.extend(self.custom_entries)

        entries = process_entries(
            entries,
            temperature=self.temperature,
            e_above_hull=self.e_above_hull,
            include_nist_data=self.include_nist_data,
            include_freed_data=self.include_freed_data,
            filter_at_temperature=self.filter_at_temperature,
            include_polymorphs=self.include_polymorphs,
            formulas_to_include=self.formulas_to_include,
            calculate_e_above_hulls=self.calculate_e_above_hulls,
            ignore_nist_solids=self.ignore_nist_solids,
        )

        doc = EntrySetDocument(
            entries=entries,
            e_above_hull=self.e_above_hull,
            include_polymorphs=self.include_polymorphs,
            formulas_to_include=self.formulas_to_include,
        )
        doc.task_label = self.name

        return doc

    @staticmethod
    def _get_exclude_elems(elems):
        """
        Get the inverse element selection. Helpful for faster querying of very large
        chemical systems).
        """
        exclude_elems = []
        for e in Element:
            if e in elems:
                continue
            exclude_elems.append(str(e))

        return exclude_elems


@dataclass
class ReactionEnumerationMaker(Maker):
    """
    Maker to create a job for enumerating reactions from a set of entries. This is
    effectively a wrapper to the run_enumerators() function (see enumerators.utils) with
    a defined output document.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.

    Args:
        name: The name of the job. This is automatically created if not provided.
    """

    name: str = "enumerate_reactions"

    @job(rxns="rxns", output_schema=EnumeratorTaskDocument)
    def make(self, enumerators: Iterable[Enumerator], entries: GibbsEntrySet):
        """
        Returns a job that enumerates reactions from a set of entries.

        NOTE: This job stores the reaction set in an additional store called
        "rxns". This needs to be configured through a user's jobflow.yaml file. See
        "additional_stores".

        Args:
            enumerators: An iterable of enumerators to perform enumeration (e.g.,
                [BasicEnumerator, BasicOpenEnumerator])
            entries: An entry set provided to each of the enumerators
        """
        data = {}

        logger.info("Running enumerators...")
        data["rxns"] = run_enumerators(enumerators, entries)
        data.update(self._get_metadata(enumerators, entries))

        logger.info("Building task document...")
        enumerator_task = EnumeratorTaskDocument(**data)
        enumerator_task.task_label = self.name
        return enumerator_task

    def _get_metadata(self, enumerators: Iterable[Enumerator], entries: GibbsEntrySet):
        """Acquires common metadata provided to output document."""
        chemsys = "-".join(entries.chemsys)
        targets = {
            target for enumerator in enumerators for target in enumerator.targets
        }

        added_elements = None
        added_chemsys = None

        if targets:
            added_elements, added_chemsys = get_added_elem_data(entries, targets)

        metadata = {
            "elements": [Element(e) for e in chemsys.split("-")],
            "chemsys": chemsys,
            "enumerators": [e.as_dict() for e in enumerators],
            "targets": list(sorted(targets)),
            "added_elements": added_elements,
            "added_chemsys": added_chemsys,
        }
        return metadata


@dataclass
class CalculateCompetitionMaker(Maker):
    """
    Maker to create job for calculating selectivities of a set of reactions given a
    provided target formula. This is a component of the SynthesisPlanningFlowMaker.

    If you use this code in your work, please consider citing the following work:

        McDermott, M. J.; McBride, B. C.; Regier, C.; Tran, G. T.; Chen, Y.; Corrao, A.
        A.; Gallant, M. C.; Kamm, G. E.; Bartel, C. J.; Chapman, K. W.; Khalifah, P. G.;
        Ceder, G.; Neilson, J. R.; Persson, K. A. Assessing Thermodynamic Selectivity of
        Solid-State Reactions for the Predictive Synthesis of Inorganic Materials. arXiv
        August 22, 2023. https://doi.org/10.48550/arXiv.2308.11816.

    Args:
        name: The name of the job. Automatically created if not provided.
        open_elem: An optional open element for performing selectivity calculations
        chempot: A (relative) chemical potential of the open element, if any. Defaults
            to 0 eV/atom.
        calculate_competition: Whether or not to calculate competition scores for
            reactions. See PrimaryCompetitionCalculator and
            SecondaryCompetitionCalculator. Defaults to True.
        calculate_chempot_distances: Whether or not to calculate chemical potential
            distances for reactions. See ChempotDistanceCalculator. Defaults to True.
        chunk_size: The number of reactions to put into each parallelized chunk. See
            class variable, CHUNK_SIZE. This will automatically be re-computed if
            out-of-memory issues are anticipated.
        cpd_kwargs: Optional keyword arguments passed to ChempotDistanceCalculator.
    """

    CHUNK_SIZE = 100

    name: str = "calculate_competition"
    open_elem: Element | None = None
    chempot: float = 0.0
    calculate_competition: bool = True
    calculate_chempot_distances: bool = True
    chunk_size: int = CHUNK_SIZE
    cpd_kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        self.open_elem = Element(self.open_elem) if self.open_elem else None
        self.open_formula = (
            Composition(str(self.open_elem)).reduced_formula if self.open_elem else None
        )

    @job(rxns="rxns", output_schema=CompetitionTaskDocument)
    def make(
        self, rxn_sets: list[ReactionSet], entries: GibbsEntrySet, target_formula: str
    ):
        """
        Returns a job that calculates competition scores and/or chemical potential
        distances for all synthesis reactions to a target phase given a provided list of
        reaction sets.

        NOTE: This job stores the reaction set in an additional store called
        "rxns". This needs to be configured through a user's jobflow.yaml file. See
        "additional_stores".

        Args:
            rxn_sets: a list of reaction sets making up all enumerated reactions in the
                chemical reaction network of interest. These will automatically be
                combined and reprocessed to match the specified conditions (open_elem +
                chempot).
            entries: The entry set used to enumerate all provided reactions. This will
                be used to facilitate selectivity calculations and ensure all reaction
                sets can be easily combined.
            target_formula: The formula of the desired target phase. This will be used
                to identify all synthesis reactions (i.e., those that produce the
                target).

        Returns:
            A job that returns synthesis reactions to the target phase, decorated with
            the relevant selectivity metrics.
        """
        if not ray.is_initialized():
            initialize_ray()

        target_formula = Composition(target_formula).reduced_formula
        added_elements, added_chemsys = get_added_elem_data(entries, [target_formula])

        logger.info("Loading reactions..")
        all_rxns = rxn_sets[0]
        for rxn_set in rxn_sets[1:]:
            all_rxns = all_rxns.add_rxn_set(rxn_set)

        if self.open_elem:  # reinitialize with open element
            all_rxns = ReactionSet(
                all_rxns.entries,
                all_rxns.indices,
                all_rxns.coeffs,
                self.open_elem,
                self.chempot,
                all_rxns.all_data,
            )

        size = len(all_rxns)  # need to get size before storing in ray

        logger.info("Identifying target reactions...")

        target_rxns = all_rxns.get_rxns_by_product(target_formula, return_set=True)

        logger.info(
            f"Identified {len(target_rxns)} target reactions out of"
            f" {size} total reactions."
        )

        logger.info(
            "Removing unnecessary reactions from total reactions to save memory..."
        )

        all_target_reactants = {
            reactant.reduced_formula for r in target_rxns for reactant in r.reactants
        }
        all_rxns = all_rxns.get_rxns_by_reactants(
            all_target_reactants, return_set=True  # type: ignore
        )

        logger.info(f"Keeping {len(all_rxns)} out of {size} total reactions...")

        size = len(all_rxns)

        logger.info("Placing reactions in ray object store...")

        all_rxns = ray.put(all_rxns.as_dict())
        logger.info("Beginning competition calculations...")

        decorated_rxns = target_rxns

        if self.calculate_competition:
            decorated_rxns = self._get_competition_decorated_rxns(
                target_rxns, all_rxns, size
            )

        logger.info("Calculating chemical potential distances...")

        if self.calculate_chempot_distances:
            decorated_rxns = self._get_chempot_decorated_rxns(decorated_rxns, entries)

        logger.info("Saving decorated reactions.")
        results = ReactionSet.from_rxns(decorated_rxns, entries=entries)

        data = {
            "rxns": results,
            "target_formula": target_formula,
            "open_elem": self.open_elem,
            "chempot": self.chempot,
            "added_elements": added_elements,
            "added_chemsys": added_chemsys,
            "calculate_competition": self.calculate_competition,
            "calculate_chempot_distances": self.calculate_chempot_distances,
            "cpd_kwargs": self.cpd_kwargs,
        }

        doc = CompetitionTaskDocument(**data)
        doc.task_label = self.name
        return doc

    def _get_competition_decorated_rxns(self, target_rxns, all_rxns, num_rxns):
        """Parallelized calculation of competition scores."""
        rxn_chunk_refs = []

        memory_per_rxn = 350  # approx 300 bytes overhead per rxn (conservative)

        num_cpus = ray.cluster_resources()["CPU"]
        available_memory = ray.cluster_resources()["memory"]

        chunk_size = self.chunk_size
        needed_memory_per_cpu = memory_per_rxn * num_rxns

        if num_cpus * needed_memory_per_cpu > available_memory:
            logger.info(
                "Not enough memory to use all CPUs simultaneously. Adjusting...."
            )
            chunk_size = (
                int(len(target_rxns) // (available_memory // needed_memory_per_cpu)) + 1
            )
            logger.info(f"Setting new chunk size to {chunk_size}.")

        for chunk in grouper(
            target_rxns,
            chunk_size,
            fillvalue=None,
        ):
            rxn_chunk_refs.append(
                _get_competition_decorated_rxns_by_chunk.options(
                    memory=num_rxns * memory_per_rxn
                ).remote(chunk, all_rxns, self.open_formula)
            )

        decorated_rxns = []
        size = len(rxn_chunk_refs)

        for r_set in tqdm(
            to_iterator(rxn_chunk_refs),
            desc="Calculating competition...",
            total=size,
        ):
            decorated_rxns.extend(r_set)

        return ReactionSet.from_rxns(decorated_rxns)

    def _get_chempot_decorated_rxns(self, target_rxns, entries):
        """Parallelized calculation of chemical potential distances."""
        if not ray.is_initialized():
            initialize_ray()

        rxn_chunk_refs = []

        open_elem = target_rxns.open_elem
        chempot = target_rxns.chempot

        entries = ray.put(entries)
        cpd_kwargs = ray.put(self.cpd_kwargs)

        for chunk in grouper(
            sorted(target_rxns, key=lambda r: r.chemical_system),
            self.chunk_size,
            fillvalue=None,
        ):
            rxn_chunk_refs.append(
                _get_chempot_decorated_rxns_by_chunk.remote(
                    chunk, entries, cpd_kwargs, open_elem, chempot
                )
            )

        size = len(rxn_chunk_refs)
        decorated_rxns = []
        for r_set in tqdm(
            to_iterator(rxn_chunk_refs),
            desc="Calculating chemical potential distances...",
            total=size,
        ):
            decorated_rxns.extend(r_set)

        return ReactionSet.from_rxns(decorated_rxns)


@dataclass
class NetworkMaker(Maker):
    """
    Maker for generating reaction networks and performing pathfinding from a set of
    previously enumerated reactions. This is a component of the NetworkFlowMaker.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.

    Args:
        name: The name of the job. Automatically selected if not provided.
        cost_function: The cost function used to calculate the edge weights in the
            network.
        precursors: An optional list of precursor formulas used in call to
            set_precursors().
        targets: An optional list of target formulas used in calls to set_target() as
            part of pathfinding.
        calculate_pathways: If integer is provided, basic pathfinding is performed using
            the find_pathways() function up to the provided number of paths (k).
            Defaults to k=10. Provide a value of None if no pathfinding is to be
            performed.
        open_elem: An optional open element used to re-initialize the total reaction
            set.
        chempot: The chemical potential of the open element, if any. Defaults to 0.
    """

    name: str = "build_and_analyze_network"
    cost_function: CostFunction = field(default_factory=Softplus)
    precursors: list[str] | None = None
    targets: list[str] | None = None
    calculate_pathways: int | None = 10
    open_elem: Element | None = None
    chempot: float = 0.0

    @job(network="network", output_schema=NetworkTaskDocument)
    def make(
        self,
        rxn_sets: Iterable[ReactionSet],
    ):
        """
        Returns a job that creates a reaction network from an iterable of reaction sets
        (coming from reaction enumeration).

        NOTE: This job stores the network object in an additional store called
        "network". This needs to be configured through a user's jobflow.yaml file. See
        "additional_stores".

        Args:
            rxn_sets: An iterable of reaction sets produced by reaction enumeration.
                These are combined, re-initialized according to provided environmental
                conditions, and then used to build the network.

        Returns:
            A job that builds and stores the reaction network for enumerated reactions.
            Also performs basic pathfinding (but no balancing; see PathwaySolverMaker).

        """
        all_rxns = ReactionSet.from_rxns(
            [rxn for rxn_set in rxn_sets for rxn in rxn_set.get_rxns()],
            open_elem=self.open_elem,
            chempot=self.chempot,
        )
        all_rxns = all_rxns.filter_duplicates()

        rn = ReactionNetwork(all_rxns, cost_function=self.cost_function)
        rn.build()

        paths = None
        if self.precursors:
            rn.set_precursors(self.precursors)
        if self.targets:
            rn.set_target(self.targets[0])
        if self.calculate_pathways and self.targets:
            paths = rn.find_pathways(self.targets, k=self.calculate_pathways)

        data = {
            "network": rn,
            "paths": paths,
            "k": self.calculate_pathways,
            "precursors": self.precursors,
            "targets": self.targets,
        }
        doc = NetworkTaskDocument(**data)
        doc.task_label = self.name
        return doc


@dataclass
class PathwaySolverMaker(Maker):
    """
    Maker for solving balanced reaction pathways from a set of (unbalanced) pathways
    returned as part of reaction network analysis.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.

    Args:
        name: The name of the job. Automatically selected if not provided.
        cost_function: The cost function used to calculate the weights of any new edges
            identified during solving (e.g., for intermediate reactions).
        precursors: The precursor formulas for the overall net reaction. These must be
            provided and result in a valid net reaction (together with the targets).
        targets: The target formulas for the overall net reaction. These must be
            provided and result in a valid net reaction (together with the precursors).
        open_elem: An optional open element used by PathwaySolver.
        chempot: The chemical potential of the open element, if any. Defaults to 0.
        max_num_combos: The maximum allowable size of the balanced reaction pathway.
            At values <=5, the solver will start to take a significant amount of
            time to run.
        find_intermediate_rxns: Whether to find intermediate reactions; crucial for
            finding pathways where intermediates react together, as these reactions
            may not occur in the graph-derived pathways. Defaults to True.
        intermediate_rxn_energy_cutoff: An energy cutoff by which to filter down
            intermediate reactions. This can be useful when there are a large number
            of possible intermediates. < 0 means allow only exergonic reactions.
        use_basic_enumerator: Whether to use the BasicEnumerator to find
            intermediate reactions. Defaults to True.
        use_minimize_enumerator: Whether to use the MinimizeGibbsEnumerator to find
            intermediate reactions. Defaults to False.
        filter_interdependent: Whether or not to filter out pathways where reaction
            steps are interdependent. Defaults to True.
    """

    name: str = "solve pathways"
    cost_function: CostFunction = field(default_factory=Softplus)
    precursors: list[str] = field(default_factory=list)
    targets: list[str] = field(default_factory=list)
    open_elem: Element | None = None
    chempot: float = 0.0
    max_num_combos: int = 4
    find_intermediate_rxns: bool = True
    intermediate_rxn_energy_cutoff: float = 0.0
    use_basic_enumerator: bool = True
    use_minimize_enumerator: bool = False
    filter_interdependent: bool = True

    def __post_init__(self):
        net_rxn = BasicReaction.balance(
            [Composition(r) for r in self.precursors],
            [Composition(p) for p in self.targets],
        )
        if not net_rxn.balanced:
            raise ValueError(
                "Can not balance pathways with specified precursors/targets. Please"
                " make sure a balanced net reaction can be written from the provided"
                " reactant and product formulas!"
            )
        self.net_rxn = net_rxn

    @job(paths="paths", output_schema=PathwaySolverTaskDocument)
    def make(
        self,
        pathways: PathwaySet,
        entries: GibbsEntrySet,
    ):
        """
        Returns a job that produces balanced pathways from raw pathways coming out of
        network pathfinding.

        NOTE: This job stores the network object in an additional store called
        "paths". This needs to be configured through a user's jobflow.yaml file. See
        "additional_stores".

        Args:
            pathways: A set of pathways (as returned by find_pathways()).
            entries: The entry set to use in path solving.

        """
        chempots = None
        if self.open_elem:
            chempots = {Element(self.open_elem): self.chempot}

        ps = PathwaySolver(
            pathways=pathways,
            entries=entries,
            cost_function=self.cost_function,
            open_elem=self.open_elem,
            chempot=self.chempot,
        )

        net_rxn = get_computed_rxn(self.net_rxn, entries, chempots)

        balanced_paths = ps.solve(
            net_rxn=net_rxn,
            max_num_combos=self.max_num_combos,
            find_intermediate_rxns=self.find_intermediate_rxns,
            intermediate_rxn_energy_cutoff=self.intermediate_rxn_energy_cutoff,
            use_basic_enumerator=self.use_basic_enumerator,
            use_minimize_enumerator=self.use_minimize_enumerator,
            filter_interdependent=self.filter_interdependent,
        )
        data = {
            "solver": ps,
            "balanced_paths": balanced_paths,
            "precursors": self.precursors,
            "targets": self.targets,
            "net_rxn": net_rxn,
            "max_num_combos": self.max_num_combos,
            "find_intermediate_rxns": self.find_intermediate_rxns,
            "intermediate_rxn_energy_cutoff": self.intermediate_rxn_energy_cutoff,
            "use_basic_enumerator": self.use_basic_enumerator,
            "use_minimize_enumerator": self.use_minimize_enumerator,
            "filter_interdependent": self.filter_interdependent,
        }

        doc = PathwaySolverTaskDocument(**data)
        doc.task_label = self.name

        return doc


def _get_competition_decorated_rxn(rxn, competing_rxns, precursors_list):
    """Calculates the primary and secondary competition for a reaction and stores them
    within the data dict."""
    if len(precursors_list) == 1:
        energy = rxn.energy_per_atom
        other_energies = np.array(
            [r.energy_per_atom for r in competing_rxns if r != rxn]
        )
        if not np.isclose(other_energies, 0.0).any():
            other_energies = np.append(other_energies, 0.0)  # need identity reaction

        primary_competition = energy - other_energies.min()

        energy_diffs = rxn.energy_per_atom - other_energies
        secondary_rxn_energies = energy_diffs[energy_diffs > 0]
        secondary_competition = (
            secondary_rxn_energies.max() if secondary_rxn_energies.any() else 0.0
        )
        rxn.data["primary_competition"] = round(primary_competition, 4)
        rxn.data["secondary_competition"] = round(secondary_competition, 4)

        decorated_rxn = rxn
    else:
        irh = InterfaceReactionHull(
            precursors_list[0],
            precursors_list[1],
            list(competing_rxns),
        )

        calc_1 = PrimaryCompetitionCalculator(irh=irh)
        calc_2 = SecondaryCompetitionCalculator(irh=irh)

        decorated_rxn = calc_1.decorate(rxn)
        decorated_rxn = calc_2.decorate(decorated_rxn)

    return decorated_rxn


@ray.remote
def _get_competition_decorated_rxns_by_chunk(rxn_chunk, all_rxns, open_formula):
    """
    Performs competition score calculations within a chunk of reactions.
    """
    decorated_rxns = []
    all_rxns = ReactionSet.from_dict(
        all_rxns
    )  # stored in Ray as a dict (should be memory efficient)

    for rxn in rxn_chunk:
        if not rxn:
            continue

        reactant_formulas = [r.reduced_formula for r in rxn.reactants]
        reactants_with_open = reactant_formulas.copy()

        if open_formula:
            reactants_with_open.append(open_formula)

        competing_rxns = all_rxns.get_rxns_by_reactants(reactants_with_open)

        if len(reactant_formulas) >= 3:
            if open_formula:
                reactant_formulas = list(set(reactant_formulas) - {open_formula})
            else:
                raise ValueError("Can only have 2 precursors, excluding open element!")

        decorated_rxns.append(
            _get_competition_decorated_rxn(rxn, competing_rxns, reactant_formulas)
        )

    return decorated_rxns


@ray.remote
def _get_chempot_decorated_rxns_by_chunk(
    rxn_chunk, entries, cpd_kwargs, open_elem, chempot
):
    """Calculates total chemical potential distance for a chunk of reactions."""
    cpd_calc_dict = {}
    new_rxns = []

    if open_elem:
        open_elem_set = {open_elem}

    rxn_chunk = [r for r in rxn_chunk if r is not None]

    for rxn in sorted(rxn_chunk, key=lambda rxn: len(rxn.elements), reverse=True):
        chemsys = rxn.chemical_system
        elems = chemsys.split("-")

        for c, cpd_calc in cpd_calc_dict.items():
            if set(c.split("-")).issuperset(elems):
                break
        else:
            if open_elem:
                filtered_entries = entries.get_subset_in_chemsys(
                    elems + [str(open_elem)]
                )
                filtered_entries = [
                    e.to_grand_entry({Element(open_elem): chempot})
                    for e in filtered_entries
                    if set(e.composition.elements) != open_elem_set
                ]
            else:
                filtered_entries = entries.get_subset_in_chemsys(elems)

            cpd_calc = ChempotDistanceCalculator.from_entries(
                filtered_entries, **cpd_kwargs
            )
            cpd_calc_dict[chemsys] = cpd_calc

        new_rxn = cpd_calc.decorate(rxn)
        new_rxns.append(new_rxn)

    results = ReactionSet.from_rxns(new_rxns, entries=entries)
    return results



"""Functions for performing Pareto front analysis."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from pandas import DataFrame


def get_pareto_front(
    df: DataFrame,
    metrics: tuple[str, ...] = (
        "energy",
        "primary_competition",
        "secondary_competition",
    ),
    maximize: bool = False,
):
    """Get the Pareto Front for a dataframe of reactions over the specified columns.

    Args:
        df: pandas DataFrame containing synthesis reactions
        metrics: Names of columns over which to calculate the Pareto front
        maximize: Whether or not maximal metrics are desired. Defaults to
            False (i.e., desired solution is to minimize the metrics).

    """
    df_original = df.copy()
    df = df_original[list(metrics)]
    pts = df.to_numpy()

    if maximize:
        pts[:, 1:] = pts[:, 1:] * -1

    return df_original[is_pareto_efficient(pts, return_mask=True)]


def is_pareto_efficient(costs, return_mask=True):
    """
    Directly lifted from @Peter's numpy-based solution on stackoverflow. Please
    give him an upvote here: https://stackoverflow.com/a/40239615. Thank you @Peter!

    Args:
        costs: An (n_points, n_costs) array
        return_mask: True to return a mask

    Returns:
        An array of indices of pareto-efficient points. If return_mask is True, this
        will be an (n_points, ) boolean array. Otherwise it will be a
        (n_efficient_points, ) integer array of indices.
    """
    is_efficient = np.arange(costs.shape[0])
    n_points = costs.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(costs):
        nondominated_point_mask = np.any(costs < costs[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        costs = costs[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
    if return_mask:
        is_efficient_mask = np.zeros(n_points, dtype=bool)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask

    return is_efficient



"""
Utility functions used by the reaction enumerator classes.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from pymatgen.analysis.interface_reactions import (
    GrandPotentialInterfacialReactivity,
    InterfacialReactivity,
)
from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.core.periodic_table import Element
from pymatgen.entries.computed_entries import ComputedEntry

from rxn_network.reactions.computed import ComputedReaction
from rxn_network.reactions.open import OpenComputedReaction
from rxn_network.utils.funcs import get_logger

if TYPE_CHECKING:
    from pymatgen.analysis.phase_diagram import GrandPotentialPhaseDiagram
    from pymatgen.entries.computed_entries import Entry

    from rxn_network.core import Composition
    from rxn_network.entries.entry_set import GibbsEntrySet
    from rxn_network.enumerators.base import Enumerator
    from rxn_network.reactions.base import Reaction

logger = get_logger(__name__)


def get_computed_rxn(
    rxn: Reaction, entries: GibbsEntrySet, chempots: dict[Element, float] | None = None
) -> ComputedReaction:
    """
    Provided with a Reaction object and a list of possible entries, this function
    returns a new ComputedReaction object containing a selection of those entries.

    Args:
        rxn: Reaction object
        entries: Iterable of entries

    Returns:
        A ComputedReaction object transformed from a normal Reaction object
    """
    reactant_entries = [
        entries.get_min_entry_by_formula(r.reduced_formula) for r in rxn.reactants
    ]
    product_entries = [
        entries.get_min_entry_by_formula(p.reduced_formula) for p in rxn.products
    ]

    if chempots:
        rxn = OpenComputedReaction.balance(reactant_entries, product_entries, chempots)
    else:
        rxn = ComputedReaction.balance(reactant_entries, product_entries)

    return rxn


def react_interface(
    r1: Composition,
    r2: Composition,
    filtered_entries: GibbsEntrySet,
    pd: PhaseDiagram,
    grand_pd: GrandPotentialPhaseDiagram | None = None,
):
    """Simple API for InterfacialReactivity module from pymatgen."""
    chempots = None

    if grand_pd:
        interface = GrandPotentialInterfacialReactivity(
            r1,
            r2,
            grand_pd,
            pd_non_grand=pd,
            norm=True,
            include_no_mixing_energy=True,
            use_hull_energy=True,
        )
        chempots = grand_pd.chempots

    else:
        interface = InterfacialReactivity(
            r1,
            r2,
            pd,
            use_hull_energy=True,
        )

    rxns = []
    for _, _, _, rxn, _ in interface.get_kinks():
        rxn = get_computed_rxn(rxn, filtered_entries, chempots)
        rxns.append(rxn)

    return rxns


def get_elems_set(entries: Iterable[Entry]) -> set[str]:
    """
    Returns chemical system as a set of element names, for set of entries.

    Args:
        entries: An iterable of entry-like objects

    Returns:
        Set of element names (strings).
    """
    return {str(elem) for e in entries for elem in e.composition.elements}


def get_total_chemsys_str(
    entries: Iterable[Entry], open_elems: Iterable[Element] | None = None
) -> str:
    """
    Returns chemical system string for set of entries, with optional open element.

    Args:
        entries: An iterable of entry-like objects
        open_elem: optional open element to include in chemical system
    """
    elements = {elem for entry in entries for elem in entry.composition.elements}
    if open_elems:
        elements.update(list(open_elems))
    return "-".join(sorted([str(e) for e in elements]))


def group_by_chemsys(
    combos: Iterable[tuple[Entry, ...]], open_elems: Iterable[Element] | None = None
) -> dict:
    """
    Groups entry combinations by chemical system, with optional open element.

    Args:
        combos: Iterable of entry combinations
        open_elem: optional open element to include in chemical system grouping

    Returns:
        Dictionary of entry combos grouped by chemical system
    """
    combo_dict: dict[str, list[tuple[Entry, ...]]] = {}
    for combo in combos:
        key = get_total_chemsys_str(combo, open_elems)
        if key in combo_dict:
            combo_dict[key].append(combo)
        else:
            combo_dict[key] = [combo]

    return combo_dict


def stabilize_entries(
    pd: PhaseDiagram, entries_to_adjust: Iterable[Entry], tol: float = 1e-6
) -> list[Entry]:
    """
    Simple method for stabilizing entries by decreasing their energy to be on the hull.

    WARNING: This method is not guaranteed to work *simultaneously* for all entries due
    to the fact that stabilization of one entry may destabilize others. Use with
    caution.

    Args:
        pd: PhaseDiagram object
        entries_to_adjust: Iterable of entries requiring energies to be adjusted
        tol: Numerical tolerance to ensure that the energy of the entry is below the
            hull

    Returns:
        A list of new entries with energies adjusted to be on the hull
    """
    indices = [pd.all_entries.index(entry) for entry in entries_to_adjust]

    new_entries = []
    for _, entry in zip(indices, entries_to_adjust):
        e_above_hull = pd.get_e_above_hull(entry)
        entry_dict = entry.to_dict()
        entry_dict["energy"] = entry.uncorrected_energy + (
            e_above_hull * entry.composition.num_atoms - tol
        )
        new_entry = ComputedEntry.from_dict(entry_dict)
        new_entries.append(new_entry)

    return new_entries


def run_enumerators(enumerators: Iterable[Enumerator], entries: GibbsEntrySet):
    """
    Utility method for calling enumerate() for a list of enumerators on a particular set
    of entries. Reaction sets are automatically combined and duplicates are filtered.

    Args:
        enumerators: an iterable of enumerators to use for reaction enumeration
        entries: an entry set to provide to the enumerate() function.
    """
    rxn_set = None
    for enumerator in enumerators:
        logger.info(f"Running {enumerator.__class__.__name__}")
        rxns = enumerator.enumerate(entries)

        logger.info(f"Adding {len(rxns)} reactions to reaction set")

        if rxn_set is None:
            rxn_set = rxns
        else:
            rxn_set = rxn_set.add_rxn_set(rxns)

    logger.info("Completed reaction enumeration. Filtering duplicates...")
    if rxn_set is not None:
        rxn_set = rxn_set.filter_duplicates()
    logger.info("Completed duplicate filtering.")
    return rxn_set



"""
Basic interface for a reaction Enumerator.
"""
from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Collection

from monty.json import MSONable

if TYPE_CHECKING:
    from rxn_network.core import Composition
    from rxn_network.entries.entry_set import GibbsEntrySet


class Enumerator(MSONable, metaclass=ABCMeta):
    """
    Base definition for a class that enumerates reactions.
    """

    def __init__(
        self,
        precursors: Collection[str | Composition] | None,
        targets: Collection[str | Composition] | None,
    ):
        self.logger = logging.getLogger(str(self.__class__.__name__))
        self.logger.setLevel("INFO")
        self.precursors = precursors or []
        self.targets = targets or []

    @abstractmethod
    def enumerate(self, entries: GibbsEntrySet):
        """
        Enumerates the potential reactions from the list of entries
        """



"""
Implements a class for conveniently and efficiently storing sets of Pathway-based
objects which share entries/reactions.
"""
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from monty.json import MSONable

from rxn_network.pathways.balanced import BalancedPathway
from rxn_network.pathways.basic import BasicPathway
from rxn_network.reactions.reaction_set import ReactionSet

if TYPE_CHECKING:
    from rxn_network.pathways.base import Pathway


class PathwaySet(MSONable):
    """
    A lightweight class for storing large sets of Pathway objects. Automatically
    represents a set of pathways as a (non-rectangular) 2D array of indices
    corresponding to reactions within a reaction set. This is useful for dumping
    reaction pathway data to a database.

    This object can easily be initialized through the from_paths() constructor.
    """

    def __init__(
        self,
        reaction_set: ReactionSet,
        indices: np.ndarray | list[list[int]],
        coefficients: np.ndarray | list[list[float] | None],
        costs: np.ndarray | list[list[float]],
    ):
        """
        Args:
            reaction_set: The reaction set containing all reactions in the pathways.
            indices: A list of lists of indices corresponding to reactions in the
                reaction set.
            coefficients: An array or list of coefficients representing the
                multiplicities (i.e., how much of) each reaction in the pathway.
            costs: An array or list of costs for each pathway.
        """
        self.reaction_set = reaction_set
        self.indices = indices
        self.coefficients = coefficients
        self.costs = costs

    @cached_property
    def paths(self) -> list[Pathway]:
        """
        Returns list of Pathway objects represented by the PathwaySet. Cached for
        efficiency.
        """
        return self._get_paths()

    def _get_paths(
        self,
    ) -> list[Pathway]:
        """
        Returns list of Pathway objects represented by the PathwaySet.
        """
        paths = []

        rxns = list(self.reaction_set.get_rxns())

        for indices, coefficients, costs in zip(
            self.indices,
            self.coefficients,
            self.costs,
        ):
            reactions = [rxns[i] for i in indices]
            if coefficients is not None:
                path = BalancedPathway(
                    reactions=reactions, coefficients=coefficients, costs=costs
                )

            else:
                path = BasicPathway(reactions=reactions, costs=costs)

            paths.append(path)

        return paths

    @classmethod
    def from_paths(
        cls,
        paths: list[Pathway],
    ) -> PathwaySet:
        """
        Initialize a PathwaySet object from a list of paths.

        Args:
            paths: List of Pathway objects
        """
        indices, coefficients, costs = [], [], []

        reaction_set = cls._get_reaction_set(paths)
        rxns = list(reaction_set.get_rxns())

        for path in paths:
            indices.append([rxns.index(r) for r in path.reactions])
            coefficients.append(getattr(path, "coefficients", None))
            costs.append(path.costs)

        return cls(
            reaction_set=reaction_set,
            indices=indices,
            coefficients=coefficients,
            costs=costs,
        )

    @staticmethod
    def _get_reaction_set(
        paths: list[Pathway],
    ) -> ReactionSet:
        """
        Returns a reaction set built from a list of paths.
        """
        return ReactionSet.from_rxns([rxn for path in paths for rxn in path.reactions])

    def __iter__(self):
        """
        Iterates over the PathwaySet.
        """
        return iter(self.paths)

    def __len__(self) -> int:
        """
        Returns the number of pathways in the PathwaySet.
        """
        return len(self.indices)



"""Core flows for the reaction-network package."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Collection

from jobflow import Flow, Maker
from pymatgen.core.composition import Element

from rxn_network.core import Composition
from rxn_network.enumerators.basic import BasicEnumerator, BasicOpenEnumerator
from rxn_network.enumerators.minimize import (
    MinimizeGibbsEnumerator,
    MinimizeGrandPotentialEnumerator,
)
from rxn_network.jobs.core import (
    CalculateCompetitionMaker,
    GetEntrySetMaker,
    NetworkMaker,
    PathwaySolverMaker,
    ReactionEnumerationMaker,
)
from rxn_network.utils.funcs import get_logger

if TYPE_CHECKING:
    from rxn_network.entries.entry_set import GibbsEntrySet

logger = get_logger(__name__)


@dataclass
class SynthesisPlanningFlowMaker(Maker):
    """
    Maker to create an inorganic synthesis planning workflow. This flow has three
    stages:

    1)  Entries are acquired via `GetEntrySetMaker`. This job both gets the computed
        entries from a databse (e.g., Materials Project) and processes them for use in
        the reaction network.
    2)  Reactions are enumerated via the provided `ReactionEnumerationMaker` (and
        associated enumerators). This computes the full reaction network so that
        selectivities can be calculated.
    3)  The competition of all synthesis reactions to the desired target is assessed via
        the `CalculateCompetitionMaker`.

    This flow also has the option to include an "open" element and a list of chempots.
    This will enumerate reactions at different conditions and evaluate their
    selectivities at those conditinons.

    This flow does not produce a specific output document. Instead, it is convenient to
    analyze output documents from each of the jobs in the flow based on the desired
    analysis. For the final "results", one should access the reaction set produced by
    the `CalculateCompetitionMaker` at the conditions of interest.

    If you use this code in your work, please consider citing the following work:

        McDermott, M. J.; McBride, B. C.; Regier, C.; Tran, G. T.; Chen, Y.; Corrao, A.
        A.; Gallant, M. C.; Kamm, G. E.; Bartel, C. J.; Chapman, K. W.; Khalifah, P. G.;
        Ceder, G.; Neilson, J. R.; Persson, K. A. Assessing Thermodynamic Selectivity of
        Solid-State Reactions for the Predictive Synthesis of Inorganic Materials. arXiv
        August 22, 2023. https://doi.org/10.48550/arXiv.2308.11816.

    Args:
        name: Name of the flow. Automatically generated if not provided.
        get_entry_set_maker: `GetEntrySetMaker`used to create the job for acquiring
            entries. Automatically generated with default settings if not provided.
        enumeration_maker: `ReactionEnumerationMaker` used to create the reaction
            enumeration job. Automatically generated with default settings if not
            provided.
        calculate_competition_maker: `CalculateCompetitionMaker` used to create the
            selectivity analysis job. Automatically generated with default settings if
            not provided.
        open_elem: Optional element to use as the "open" element. If provided, the flow
            will  enumerate reactions at different chemical potentials of this element.
        chempots: List of chemical potentials to use for the "open" element. If
            provided, the flow will enumerate reactions at different chemical potentials
            of this element.
        use_basic_enumerators: Whether to use the `BasicEnumerator` and
            `BasicOpenEnumerator` enumerators in the enumeration job.
        use_minimize_enumerators: Whether to use the `MinimizeGibbsEnumerator` and the
            `MinimizeGrandPotentialEnumerator` enumerators in the enumeration job.
        basic_enumerator_kwargs: Keyword arguments to pass to the basic enumerators.
        minimize_enumerator_kwargs: Keyword arguments to pass to the minimize
            enumerators.
    """

    name: str = "synthesis_planning"
    get_entry_set_maker: GetEntrySetMaker = field(default_factory=GetEntrySetMaker)
    enumeration_maker: ReactionEnumerationMaker = field(
        default_factory=ReactionEnumerationMaker
    )
    calculate_competition_maker: CalculateCompetitionMaker = field(
        default_factory=CalculateCompetitionMaker
    )
    open_elem: Element | str | None = None
    chempots: list[float] | None = None
    use_basic_enumerators: bool = True
    use_minimize_enumerators: bool = True
    basic_enumerator_kwargs: dict = field(default_factory=dict)
    minimize_enumerator_kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        self.open_elem = Element(self.open_elem) if self.open_elem else None
        self.open_formula = (
            Composition(str(self.open_elem)).reduced_formula if self.open_elem else None
        )

    def make(  # type: ignore
        self,
        target_formula: str,
        added_elems: Collection[str] | None = None,
        entries: GibbsEntrySet | None = None,
    ):
        """
        Returns a flow used for planning optimal synthesis recipes to a specified
        target.

        Args:
            target_formula: The chemical formula of a target phase (e.g., "BaTiO3").
            added_elems: An optional list of additional elements to consider (e.g.,
                ["C", "H"]). Defaults to None.
            entries: An optional provided set of entries to enumerate from. If entries
                are not provided, then they will be acquired from a database (e.g.,
                Materials Project) and processed using the GetEntrySetMaker.

        """
        target_formula = Composition(target_formula).reduced_formula

        flow_name = f"Synthesis planning: {target_formula}"

        if added_elems is None:
            added_elems = []
        else:
            flow_name = flow_name + f" (+ {'-'.join(sorted(added_elems))})"

        flow_name = flow_name + f", T={self.get_entry_set_maker.temperature} K"

        chemsys = "-".join(
            sorted(
                {str(e) for e in Composition(target_formula).elements}
                | {str(e) for e in added_elems}
            )
        )

        jobs = []

        if entries is None:
            get_entry_set_maker = self.get_entry_set_maker.update_kwargs(
                {
                    "name": self.get_entry_set_maker.name
                    + f" ({chemsys}, T={self.get_entry_set_maker.temperature} K,"
                    f" +{round(self.get_entry_set_maker.e_above_hull, 3)} eV)",
                    "formulas_to_include": list(
                        set(
                            self.get_entry_set_maker.formulas_to_include
                            + [target_formula]
                        )
                    ),
                }
            )
            get_entry_set_job = get_entry_set_maker.make(chemsys)
            jobs.append(get_entry_set_job)
            entries = get_entry_set_job.output.entries

        targets = [target_formula]
        filter_by_chemsys = Composition(target_formula).chemical_system

        basic_enumerator_kwargs = self.basic_enumerator_kwargs.copy()
        minimize_enumerator_kwargs = self.minimize_enumerator_kwargs.copy()

        kwarg_update = {"targets": targets, "filter_by_chemsys": filter_by_chemsys}

        basic_enumerator_kwargs.update(kwarg_update)
        minimize_enumerator_kwargs.update(kwarg_update)

        enumerators = []

        if self.use_basic_enumerators:
            enumerators.append(
                BasicEnumerator(
                    filter_by_chemsys=filter_by_chemsys,
                    **self.basic_enumerator_kwargs,
                )
            )
        if self.use_minimize_enumerators:
            enumerators.append(
                MinimizeGibbsEnumerator(
                    filter_by_chemsys=filter_by_chemsys,
                    **self.minimize_enumerator_kwargs,
                )
            )

        enumeration_job = self.enumeration_maker.make(
            enumerators=enumerators, entries=entries
        )
        jobs.append(enumeration_job)

        base_rxn_set = enumeration_job.output.rxns
        calculate_competition_maker = self.calculate_competition_maker

        base_calculate_competition_job = calculate_competition_maker.make(
            rxn_sets=[base_rxn_set],
            entries=entries,
            target_formula=target_formula,
        )
        jobs.append(base_calculate_competition_job)

        if self.open_elem and self.chempots:
            for chempot in self.chempots:
                subname = f"(open {str(self.open_elem)}, mu={chempot})"
                enumeration_maker = self.enumeration_maker.update_kwargs(
                    {"name": self.enumeration_maker.name + subname},
                    nested=False,
                )
                calculate_competition_maker = calculate_competition_maker.update_kwargs(
                    {
                        "chempot": chempot,
                        "open_elem": self.open_elem,
                        "name": self.calculate_competition_maker.name + subname,
                    },
                    nested=False,
                )

                open_enumerators = []

                if self.use_basic_enumerators:
                    open_enumerators.append(
                        BasicOpenEnumerator(
                            open_phases=[self.open_formula],
                            **self.basic_enumerator_kwargs,
                        )
                    )
                if self.use_minimize_enumerators:
                    open_enumerators.append(
                        MinimizeGrandPotentialEnumerator(
                            open_elem=self.open_elem,
                            mu=chempot,
                            filter_by_chemsys=filter_by_chemsys,
                            **self.minimize_enumerator_kwargs,
                        )
                    )
                enumeration_job = enumeration_maker.make(
                    enumerators=open_enumerators,
                    entries=entries,
                )
                jobs.append(enumeration_job)

                calculate_competition_job = calculate_competition_maker.make(
                    rxn_sets=[base_rxn_set, enumeration_job.output.rxns],
                    entries=entries,
                    target_formula=target_formula,
                )

                jobs.append(calculate_competition_job)

        return Flow(jobs, name=flow_name)


@dataclass
class NetworkFlowMaker(Maker):
    """
    Maker to create a chemical reaction network and perform (balanced) pathfinding on
    the network.

    This flow has four stages:

    1)  Entries are acquired via `GetEntrySetMaker`. This job both gets the computed
        entries from a databse (e.g., Materials Project) and processes them for use in
        the reaction network.
    2)  Reactions are enumerated via the provided `ReactionEnumerationMaker` (and
        associated enumerators).
    3)  The network is created using `NetworkMaker` and basic paths are found
        (k-shortest paths to each target).
    4)  The final balanced reaction pathways are produced using the `SolverMaker`.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.

    Args:
        name: The name of the network flow. Automatically assigned if not provided.
        get_entry_set_maker: `GetEntrySetMaker`used to create the job for acquiring
            entries. Automatically generated with default settings if not provided.
        enumeration_maker: `ReactionEnumerationMaker` used to create the reaction
            enumeration job. Automatically generated with default settings if not
            provided.
        network_maker: `NetworkMaker` used to create the reaction network from sets of
            reactions. Also identifies basic reaction pathways. Automatically generated
            with default settings if not provided.
        solver_maker: `PathwaySolverMaker` used to find balanced reaction pathways from
            set of pathways emerging from pathfinding. Automatically generated with
            default settings if not provided.
        open_elem: Optional element to use as the "open" element. If provided, the flow
            will enumerate reactions at different chemical potentials of this element.
        chempots: List of chemical potentials to use for the "open" element. If
            provided, the flow will enumerate reactions at different chemical potentials
            of this element.
        use_basic_enumerators: Whether to use the `BasicEnumerator` and
            `BasicOpenEnumerator` enumerators in the enumeration job.
        use_minimize_enumerators: Whether to use the `MinimizeGibbsEnumerator` and the
            `MinimizeGrandPotentialEnumerator` enumerators in the enumeration job.
        basic_enumerator_kwargs: Keyword arguments to pass to the basic enumerators.
        minimize_enumerator_kwargs: Keyword arguments to pass to the minimize
            enumerators.
    """

    name: str = "find_reaction_pathways"
    get_entry_set_maker: GetEntrySetMaker = field(default_factory=GetEntrySetMaker)
    enumeration_maker: ReactionEnumerationMaker = field(
        default_factory=ReactionEnumerationMaker
    )
    network_maker: NetworkMaker = field(default_factory=NetworkMaker)
    solver_maker: PathwaySolverMaker | None = None
    open_elem: Element | None = None
    chempots: list[float] | None = None
    use_basic_enumerators: bool = True
    use_minimize_enumerators: bool = True
    basic_enumerator_kwargs: dict = field(default_factory=dict)
    minimize_enumerator_kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        self.open_elem = Element(self.open_elem) if self.open_elem else None
        self.open_formula = (
            Composition(str(self.open_elem)).reduced_formula if self.open_elem else None
        )

    def make(self, precursors, targets, entries=None):
        precursor_formulas = [Composition(f).reduced_formula for f in precursors]
        target_formulas = [Composition(f).reduced_formula for f in targets]

        flow_name = (
            f"Reaction Network analysis: {'-'.join(sorted(precursor_formulas))} ->"
            f" {'-'.join(sorted(target_formulas))}"
        )
        chemsys = "-".join(
            {
                str(e)
                for formula in precursor_formulas + target_formulas
                for e in Composition(formula).elements
            }
        )

        jobs = []

        if entries is None:
            get_entry_set_maker = self.get_entry_set_maker.update_kwargs(
                {
                    "name": self.get_entry_set_maker.name
                    + f" ({chemsys}, T={self.get_entry_set_maker.temperature} K,"
                    f" +{round(self.get_entry_set_maker.e_above_hull, 3)} eV)",
                    "formulas_to_include": list(
                        set(
                            self.get_entry_set_maker.formulas_to_include
                            + precursor_formulas
                            + target_formulas
                        )
                    ),
                }
            )
            get_entry_set_job = get_entry_set_maker.make(chemsys)
            jobs.append(get_entry_set_job)
            entries = get_entry_set_job.output.entries

        self.basic_enumerator_kwargs.copy()
        self.minimize_enumerator_kwargs.copy()

        enumerators = []

        if self.use_basic_enumerators:
            enumerators.append(
                BasicEnumerator(
                    **self.basic_enumerator_kwargs,
                )
            )
            if self.open_formula:
                enumerators.append(
                    BasicOpenEnumerator(
                        open_phases=[self.open_formula],
                        **self.basic_enumerator_kwargs,
                    )
                )
        if self.use_minimize_enumerators:
            enumerators.append(
                MinimizeGibbsEnumerator(
                    **self.minimize_enumerator_kwargs,
                )
            )

        enumeration_job = self.enumeration_maker.make(
            enumerators=enumerators, entries=entries
        )
        jobs.append(enumeration_job)

        base_rxn_set = enumeration_job.output.rxns

        base_network_job = self.network_maker.make([base_rxn_set])
        jobs.append(base_network_job)

        if self.solver_maker:
            base_pathway_job = self.solver_maker.make(
                base_network_job.output.paths,
                entries=base_network_job.output.network.entries,
            )
            jobs.append(base_pathway_job)

        if self.use_minimize_enumerators and self.open_elem and self.chempots:
            for chempot in self.chempots:
                subname = f"(open {str(self.open_elem)}, mu={chempot})"
                enumeration_maker = self.enumeration_maker.update_kwargs(
                    {"name": self.enumeration_maker.name + subname},
                    nested=False,
                )
                network_maker = self.network_maker.update_kwargs(
                    {
                        "name": self.network_maker.name + subname,
                        "chempot": chempot,
                        "open_elem": self.open_elem,
                    },
                    nested=False,
                )
                if self.solver_maker:
                    solver_maker = self.solver_maker.update_kwargs(
                        {
                            "name": self.solver_maker.name + subname,
                            "chempot": chempot,
                            "open_elem": self.open_elem,
                        },
                        nested=False,
                    )

                enumerator = MinimizeGrandPotentialEnumerator(
                    open_elem=self.open_elem,
                    mu=chempot,
                    **self.minimize_enumerator_kwargs,
                )
                enumeration_job = enumeration_maker.make(
                    enumerators=[enumerator], entries=entries
                )
                network_job = network_maker.make(
                    [base_rxn_set, enumeration_job.output.rxns]
                )
                jobs.extend([enumeration_job, network_job])
                if self.solver_maker:
                    pathway_job = solver_maker.make(
                        network_job.output.paths,
                        entries=network_job.output.network.entries,
                    )
                    jobs.append(pathway_job)

        return Flow(jobs, name=flow_name)



"""Functions for working with Ray (parallelization library)."""
import os

import ray

from rxn_network.utils.funcs import get_logger

logger = get_logger(__name__)


def initialize_ray(quiet: bool = False):
    """
    Simple function to initialize ray. Basic support for running ray on multiple nodes.
    Currently supports SLURM and PBS job schedulers.

    SLURM:
        Checks enviornment for existence of "ip_head" for situations where the user is
        running on multiple nodes. Automatically creats a new ray cluster if it has not
        been initialized. See https://github.com/NERSC/slurm-ray-cluster/
    PBS:
        Checks environment for PBS_NNODES > 1.

    """
    if not quiet:
        logger.setLevel("INFO")
    if not ray.is_initialized():
        logger.info("Ray is not initialized. Checking for existing cluster...")
        if os.environ.get("ip_head") or int(os.environ.get("PBS_NNODES", 0)) > 1:
            ray.init(
                address="auto",
            )
        else:
            logger.info(
                "Could not identify existing Ray instance. Creating a new one..."
            )
            ray.init()

        logger.info(
            f"HOST: {ray.nodes()[0]['NodeManagerHostname']}, "
            f"Num CPUs: {ray.cluster_resources()['CPU']}, "
            f"Total Memory: {ray.cluster_resources()['memory']}"
        )
    else:
        logger.info("Ray is already initialized.")


def to_iterator(obj_ids, get_obj_ids: bool = False):
    """
    Method to convert a list of ray object ids to an iterator that can be used in a for
    loop.
    """
    while obj_ids:
        done, obj_ids = ray.wait(obj_ids)
        if get_obj_ids:
            yield done[0], ray.get(done[0])
        else:
            yield ray.get(done[0])



"""
A computed entry object for estimating the Gibbs free energy of formation. Note that
this is similar to the implementation within pymatgen, but has been refactored here to
add extra functionality.
"""
from __future__ import annotations

from copy import deepcopy
from itertools import combinations
from typing import TYPE_CHECKING

import numpy as np
from monty.json import MontyDecoder
from pymatgen.analysis.phase_diagram import GrandPotPDEntry
from pymatgen.entries.computed_entries import ComputedEntry, ConstantEnergyAdjustment
from scipy.interpolate import interp1d

from rxn_network.core import Composition
from rxn_network.data import G_ELEMS

if TYPE_CHECKING:
    from pymatgen.core.periodic_table import Element
    from pymatgen.core.structure import Structure
    from pymatgen.entries.computed_entries import EnergyAdjustment


class GibbsComputedEntry(ComputedEntry):
    """
    An extension to ComputedEntry which estimates the Gibbs free energy of formation
    of solids using energy adjustments from the machine-learned SISSO descriptor from
    Bartel et al. (2018).  Note that this is similar to the implementation within
    pymatgen, but has been refactored here to add extra functionality.

    WARNING: This descriptor only applies to solids. See entries.nist.NISTReferenceEntry
    for common gases (e.g. CO2).

    If you use this entry class in your work, please consider citing the following
    paper:

        Bartel, C. J., Millican, S. L., Deml, A. M., Rumptz, J. R., Tumas, W., Weimer,
        A. W., … Holder, A. M. (2018). Physical descriptor for the Gibbs energy of
        inorganic crystalline solids and temperature-dependent materials chemistry.
        Nature Communications, 9(1), 4168. https://doi.org/10.1038/s41467-018-06682-4.
    """

    def __init__(
        self,
        composition: Composition,
        formation_energy_per_atom: float,
        volume_per_atom: float,
        temperature: float,
        energy_adjustments: list[EnergyAdjustment] | None = None,
        parameters: dict | None = None,
        data: dict | None = None,
        entry_id: object | None = None,
    ):
        """

        A new computed entry object is returned with a supplied energy correction
        representing the difference between the formation enthalpy at T=0K and the
        Gibbs formation energy at the specified temperature.

        Args:
            composition: The composition object (pymatgen)
            formation_energy_per_atom: Calculated formation enthalpy, dH, at T = 298 K,
                normalized to the total number of atoms in the composition. NOTE: since
                this is a _formation_ energy, it must be calculated using a phase
                diagram construction.
            volume_per_atom: The total volume of the associated structure divided by its
                number of atoms.
            temperature: Temperature [K] by which to acquire dGf(T); must be selected
                from a range of [300, 2000] K. If temperature is not selected from one
                of [300, 400, 500, ... 2000 K], then free energies will be interpolated.
            energy_adjustments: Optional list of energy adjustments.
            parameters: Optional list of calculation parameters.
            data: Optional dictionary containing entry data.
            entry_id: Optional entry id, such as the entry's mpid.
        """
        composition = Composition(composition)
        self._composition = composition
        self.volume_per_atom = volume_per_atom
        num_atoms = composition.num_atoms
        if temperature < 300 or temperature > 2000:
            raise ValueError("Temperature must be selected from range: [300, 2000] K.")

        if energy_adjustments is not None:
            energy_adjustments = [
                adjustment
                for adjustment in energy_adjustments
                if adjustment.name != "Gibbs SISSO Correction"
            ]
        else:
            energy_adjustments = []

        energy_adjustments.append(
            ConstantEnergyAdjustment(
                self.gibbs_adjustment(temperature),
                uncertainty=0.05 * num_atoms,  # descriptor has ~50 meV/atom MAD
                name="Gibbs SISSO Correction",
                description=f"Gibbs correction: dGf({temperature} K) - dHf (298 K)",
            )
        )

        formation_energy = num_atoms * formation_energy_per_atom

        super().__init__(
            composition=composition,
            energy=formation_energy,
            energy_adjustments=energy_adjustments,
            parameters=parameters,
            data=data,
            entry_id=entry_id,
        )
        self._composition = composition
        self.formation_energy_per_atom = formation_energy_per_atom
        self.temperature = temperature

    def get_new_temperature(self, new_temperature: float) -> GibbsComputedEntry:
        """
        Return a copy of the GibbsComputedEntry at the new specified temperature.

        Args:
            new_temperature: The new temperature to use [K]

        Returns:
            A copy of the GibbsComputedEntry, initialized at the new specified
            temperature.
        """
        new_entry_dict = self.as_dict()
        new_entry_dict["temperature"] = new_temperature

        new_entry = self.from_dict(new_entry_dict)
        return new_entry

    def gibbs_adjustment(self, temperature: float) -> float:
        """
        Returns the difference between the predicted Gibbs formation energy and the
        formation enthalpy at 298 K, i.e., dGf(T) - dHf(298 K). Calculated using
        SISSO descriptor from Bartel et al. (2018) and elemental chemical potentials
        (FactSage).

        Units: eV (not normalized)

        Args:
            temperature: The absolute temperature [K].
        Returns:
            The correction to Gibbs free energy of formation (eV) from DFT energy.
        """
        if self._composition.is_element:
            return 0

        num_atoms = self._composition.num_atoms
        reduced_mass = self._reduced_mass(self._composition)

        return num_atoms * self._g_delta_sisso(
            self.volume_per_atom, reduced_mass, temperature
        ) - self._sum_g_i(self._composition, temperature)

    def to_grand_entry(self, chempots: dict[Element, float]):
        """
        Convert a GibbsComputedEntry to a GrandComputedEntry.

        Args:
            chempots: A dictionary of {element: chempot} pairs.

        Returns:
            A GrandComputedEntry.
        """
        return GrandPotPDEntry(self, chempots)

    def copy(self):
        return deepcopy(self)

    @staticmethod
    def _g_delta_sisso(
        volume_per_atom: float, reduced_mass: float, temp: float
    ) -> float:
        """
        G^delta as predicted by SISSO-learned descriptor from Eq. (4) in Bartel et al.
        (2018).

        Args:
            vol_per_atom: volume per atom [Å^3/atom]
            reduced_mass: reduced mass as calculated with pair-wise sum formula [amu]
            temp: Temperature [K]

        Returns:
            float: G^delta
        """

        return (
            (
                -2.48e-4 * np.log(volume_per_atom)
                - 8.94e-5 * reduced_mass / volume_per_atom
            )
            * temp
            + 0.181 * np.log(temp)
            - 0.882
        )

    @staticmethod
    def _sum_g_i(composition: Composition, temperature: float) -> float:
        """
        Sum of the stoichiometrically weighted chemical potentials [eV] of the elements
        at specified temperature, as acquired from data/elements.json.
        """
        elems = composition.get_el_amt_dict()

        if temperature % 100 > 0:
            sum_g_i = 0
            for elem, amt in elems.items():
                g_interp = interp1d(
                    [float(t) for t in G_ELEMS.keys()],
                    [g_dict[elem] for g_dict in G_ELEMS.values()],
                )
                sum_g_i += amt * g_interp(temperature)
        else:
            sum_g_i = sum(
                amt * G_ELEMS[str(temperature)][elem] for elem, amt in elems.items()
            )

        return sum_g_i

    @staticmethod
    def _reduced_mass(composition: Composition) -> float:
        """
        Reduced mass [amu] as calculated via Eq. 6 in Bartel et al. (2018),
        to be used in SISSO descriptor equation.
        """
        reduced_comp = composition.reduced_composition
        num_elems = len(reduced_comp.elements)
        elem_dict = reduced_comp.get_el_amt_dict()

        denominator = (num_elems - 1) * reduced_comp.num_atoms

        all_pairs = combinations(elem_dict.items(), 2)
        mass_sum = 0.0

        for pair in all_pairs:
            m_i = Composition(pair[0][0]).weight
            m_j = Composition(pair[1][0]).weight
            alpha_i = pair[0][1]
            alpha_j = pair[1][1]

            mass_sum += (alpha_i + alpha_j) * (m_i * m_j) / (m_i + m_j)

        reduced_mass = (1 / denominator) * mass_sum

        return reduced_mass

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        formation_energy_per_atom: float,
        temperature: float,
        **kwargs,
    ) -> GibbsComputedEntry:
        """
        Constructor method for building a GibbsComputedEntry from a structure,
        formation enthalpy, and temperature.

        Args:
            structure: Structure object (pymatgen)
            formation_energy_per_atom: Formation enthalpy at T = 298 K associated
                with structure
            temperature: Desired temperature [K] for acquiring dGf(T)
            **kwargs: Optional kwargs to be passed to GibbsComputedEntry.__init__

        Returns:
            A new GibbsComputedEntry object
        """
        composition = Composition(structure.composition)
        volume_per_atom = structure.volume / structure.num_sites
        entry = cls(
            composition=composition,
            formation_energy_per_atom=formation_energy_per_atom,
            volume_per_atom=volume_per_atom,
            temperature=temperature,
            **kwargs,
        )
        return entry

    @property
    def is_experimental(self) -> bool:
        """
        Returns True if self.data contains {"theoretical": False}. If
        theoretical is not specified but there is greater than 1 icsd_id provided,
        assumes that the presence of an icsd_id means the entry is experimental.
        """
        if "theoretical" in self.data:
            return not self.data["theoretical"]
        if "icsd_ids" in self.data:
            return len(self.data["icsd_ids"]) >= 1

        return False

    @property
    def unique_id(self) -> str:
        """
        Returns a unique ID for the entry based on its entry-id and temperature. This is
        useful because the same entry-id can be used for multiple entries at different
        temperatures.
        """
        return f"{self.entry_id}_{self.temperature}"

    def as_dict(self) -> dict:
        """Returns an MSONable dict."""
        data = super().as_dict()
        data["volume_per_atom"] = self.volume_per_atom
        data["formation_energy_per_atom"] = self.formation_energy_per_atom
        data["temperature"] = self.temperature
        return data

    @classmethod
    def from_dict(cls, d: dict) -> GibbsComputedEntry:
        """ "Returns a GibbsComputedEntry object from MSONable dictionary"""
        dec = MontyDecoder()
        entry = cls(
            composition=d["composition"],
            formation_energy_per_atom=d["formation_energy_per_atom"],
            volume_per_atom=d["volume_per_atom"],
            temperature=d["temperature"],
            energy_adjustments=dec.process_decoded(d["energy_adjustments"]),
            parameters=d["parameters"],
            data=d["data"],
            entry_id=d["entry_id"],
        )
        return entry

    def __repr__(self):
        output = [
            (
                f"GibbsComputedEntry | {self.entry_id} | {self.composition.formula} "
                f"({self.composition.reduced_formula})"
            ),
            f"Gibbs Energy ({self.temperature} K) = {self.energy:.4f}",
        ]
        return "\n".join(output)

    def __eq__(self, other):
        if not type(other) is type(self):
            return False

        if not np.isclose(self.temperature, other.temperature):
            return False

        if not np.isclose(self.energy, other.energy):
            return False

        if getattr(self, "entry_id", None) and getattr(other, "entry_id", None):
            return self.entry_id == other.entry_id

        if self.composition != other.composition:
            return False

        return True

    def __hash__(self):
        return hash(
            (
                self.composition,
                self.energy,
                self.entry_id,
                self.temperature,
            )
        )



"""
Energy correction classes for entry objects.
"""

from pymatgen.entries.computed_entries import CompositionEnergyAdjustment

CARBONATE_CORRECTION = (
    0.830  # eV per (CO3)2- anion in composition; see Jupyter NB for fitting
)


class CarbonateCorrection(CompositionEnergyAdjustment):
    """
    Supplies a carbonate correction due to systematic GGA errors in carbonate formation
    energies.

    See provided jupyter NB for fitting of the correction:
    data/fit_carbonate_correction.ipynb
    """

    def __init__(
        self, num_ions: int, carbonate_correction: float = CARBONATE_CORRECTION
    ):
        """
        Initalizes a carbonate correction object

        Args:
            num_ions: Number of carbonate ions in the composition object
            carbonate_correction: Energy correction, eV per (CO3)2- anion
        """
        self._num_ions = num_ions
        self._carbonate_correction = carbonate_correction

        super().__init__(
            adj_per_atom=carbonate_correction,
            n_atoms=num_ions,
            name="Carbonate Correction",
            description=(
                "Correction for dGf with (CO3)2- anion, as fit to MP data (300 K)."
            ),
        )

    @property
    def num_ions(self) -> int:
        """
        Number of carbonate ions ion the composition object
        """
        return self._num_ions

    @property
    def carbonate_correction(self) -> float:
        """
        Energy correction for carbonate ion, eV per (CO3)2- anion
        """
        return self._carbonate_correction



"""
Basic interface for a (chemical) Reaction
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from monty.json import MSONable

if TYPE_CHECKING:
    from numpy import ndarray
    from pymatgen.core.composition import Element

    from rxn_network.core import Composition


class Reaction(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction class.
    """

    @property
    @abstractmethod
    def reactants(self) -> list[Composition]:
        """List of reactants for this reaction"""

    @property
    @abstractmethod
    def products(self) -> list[Composition]:
        """List of products for this reaction"""

    @property
    @abstractmethod
    def coefficients(self) -> ndarray:
        """Coefficients of the reaction"""

    @property
    @abstractmethod
    def energy(self) -> float:
        """The energy of this reaction in total eV"""

    @property
    @abstractmethod
    def compositions(self) -> list[Composition]:
        """List of all compositions in the reaction"""

    @property
    def elements(self) -> list[Element]:
        """List of elements in the reaction"""
        return list(set(el for comp in self.compositions for el in comp.elements))



"""
This module implements two types of reaction enumerators using a free energy
minimization technique, with or without the option of an open entry.
"""
from __future__ import annotations

from itertools import product

from pymatgen.core.composition import Element

from rxn_network.core import Composition
from rxn_network.enumerators.basic import BasicEnumerator
from rxn_network.enumerators.utils import react_interface


class MinimizeGibbsEnumerator(BasicEnumerator):
    """
    Enumerator for finding all reactions between two reactants that are predicted by
    thermodynamics; i.e., they appear when taking the convex hull along a straight line
    connecting any two phases in G-x phase space. Identity reactions are automatically
    excluded.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.
    """

    MIN_CHUNK_SIZE = 1000
    MAX_NUM_JOBS = 1000

    def __init__(
        self,
        precursors: list[str] | None = None,
        targets: list[str] | None = None,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_duplicates: bool = False,
        filter_by_chemsys: str | None = None,
        chunk_size: int = MIN_CHUNK_SIZE,
        max_num_jobs: int = MAX_NUM_JOBS,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
        max_num_constraints: int = 1,
        quiet: bool = False,
    ):
        """
        Args:
            precursors: Optional list of precursor formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as reactants. The "exclusive_precursors" parameter allows one to tune
                whether the enumerated reactions should include ALL precursors (the
                default) or just one.
            targets: Optional list of target formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as products. The "exclusive_targets" parameter allows one to tune
                whether the enumerated reactions should include ALL targets or just one
                (the default).
            exclusive_precursors: Whether enumerated reactions are required to have
                reactants that are a subset of the provided list of precursors. If True
                (the default), this only identifies reactions with reactants selected
                from the provided precursors.
            exclusive_targets: Whether enumerated reactions are required to have
                products that are a subset of the provided list of targets. If False,
                (the default), this identifies all reactions containing at least one
                composition from the provided list of targets (and any number of
                byproducts).
            filter_duplicates: Whether to remove duplicate reactions. Defaults to False.
            filter_by_chemsys: An optional chemical system for which to filter produced
                reactions by. This ensures that all output reactions contain at least
                one element within the provided system.
            chunk_size: The minimum number of reactions per chunk procssed. Needs to be
                sufficiently large to make parallelization a cost-effective strategy.
                Defaults to MIN_CHUNK_SIZE.
            max_num_jobs: The upper limit for the number of jobs created. Defaults to
                MAX_NUM_JOBS.
            remove_unbalanced: Whether to remove reactions which are unbalanced; this is
                usually advisable. Defaults to True.
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. This is also
                advisable for ensuring that only unique reaction sets are produced.
                Defaults to True.
            max_num_constraints: The maximum number of allowable constraints enforced by
                reaction balancing. Defaults to 1 (which is usually advisable).
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """
        super().__init__(
            precursors=precursors,
            targets=targets,
            exclusive_precursors=exclusive_precursors,
            exclusive_targets=exclusive_targets,
            filter_duplicates=filter_duplicates,
            filter_by_chemsys=filter_by_chemsys,
            chunk_size=chunk_size,
            max_num_jobs=max_num_jobs,
            remove_unbalanced=remove_unbalanced,
            remove_changed=remove_changed,
            max_num_constraints=max_num_constraints,
            quiet=quiet,
        )
        self._build_pd = True

    @staticmethod
    def _react_function(
        reactants, products, filtered_entries=None, pd=None, grand_pd=None, **kwargs
    ):
        """React method for MinimizeGibbsEnumerator, which uses the interfacial reaction
        approach (see _react_interface())"""
        _ = products, kwargs  # unused

        r = list(reactants)
        r0 = r[0]

        if len(r) == 1:
            r1 = r[0]
        else:
            r1 = r[1]

        return react_interface(
            r0.composition,
            r1.composition,
            filtered_entries,
            pd,
            grand_pd,
        )

    @staticmethod
    def _get_rxn_iterable(combos, open_combos):
        """Gets the iterable used to generate reactions"""
        _ = open_combos  # unused

        return product(combos, [None])

    @staticmethod
    def _rxn_iter_length(combos, open_combos):
        _ = open_combos
        return len(combos)


class MinimizeGrandPotentialEnumerator(MinimizeGibbsEnumerator):
    """
    Enumerator for finding all reactions between two reactants and an open element
    that are predicted by thermo; i.e., they appear when taking the
    convex hull along a straight line connecting any two phases in Phi-x
    phase space. Identity reactions are excluded.

    If you use this code in your own work, please consider citing this paper:

        McDermott, M. J.; Dwaraknath, S. S.; Persson, K. A. A Graph-Based Network for
        Predicting Chemical Reaction Pathways in Solid-State Materials Synthesis. Nature
        Communications 2021, 12 (1), 3097. https://doi.org/10.1038/s41467-021-23339-x.
    """

    MIN_CHUNK_SIZE = 1000
    MAX_NUM_JOBS = 1000

    def __init__(
        self,
        open_elem: Element,
        mu: float,
        precursors: list[str] | None = None,
        targets: list[str] | None = None,
        exclusive_precursors: bool = True,
        exclusive_targets: bool = False,
        filter_duplicates: bool = False,
        filter_by_chemsys: str | None = None,
        chunk_size: int = MIN_CHUNK_SIZE,
        max_num_jobs: int = MAX_NUM_JOBS,
        remove_unbalanced: bool = True,
        remove_changed: bool = True,
        max_num_constraints: int = 1,
        quiet: bool = False,
    ):
        """
        Args:
            open_elem: The element to be considered as open.
            mu: The chemical potential of the open element (eV).
            precursors: Optional list of precursor formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as reactants. The "exclusive_precursors" parameter allows one to tune
                whether the enumerated reactions should include ALL precursors (the
                default) or just one.
            targets: Optional list of target formulas. The only reactions that
                will be enumerated are those featuring one or more of these compositions
                as products. The "exclusive_targets" parameter allows one to tune
                whether the enumerated reactions should include ALL targets or just one
                (the default).
            exclusive_precursors: Whether enumerated reactions are required to have
                reactants that are a subset of the provided list of precursors. If True
                (the default), this only identifies reactions with reactants selected
                from the provided precursors.
            exclusive_targets: Whether enumerated reactions are required to have
                products that are a subset of the provided list of targets. If False,
                (the default), this identifies all reactions containing at least one
                composition from the provided list of targets (and any number of
                byproducts).
            filter_duplicates: Whether to remove duplicate reactions. Defaults to False.
            filter_by_chemsys: An optional chemical system for which to filter produced
                reactions by. This ensures that all output reactions contain at least
                one element within the provided system.
            chunk_size: The minimum number of reactions per chunk procssed. Needs to be
                sufficiently large to make parallelization a cost-effective strategy.
                Defaults to MIN_CHUNK_SIZE.
            max_num_jobs: The upper limit for the number of jobs created. Defaults to
                MAX_NUM_JOBS.
            remove_unbalanced: Whether to remove reactions which are unbalanced; this is
                usually advisable. Defaults to True.
            remove_changed: Whether to remove reactions which can only be balanced by
                removing a reactant/product or having it change sides. This is also
                advisable for ensuring that only unique reaction sets are produced.
                Defaults to True.
            max_num_constraints: The maximum number of allowable constraints enforced by
                reaction balancing. Defaults to 1 (which is usually advisable).
            quiet: Whether to run in quiet mode (no progress bar). Defaults to False.
        """

        super().__init__(
            precursors=precursors,
            targets=targets,
            exclusive_precursors=exclusive_precursors,
            exclusive_targets=exclusive_targets,
            filter_duplicates=filter_duplicates,
            filter_by_chemsys=filter_by_chemsys,
            chunk_size=chunk_size,
            max_num_jobs=max_num_jobs,
            remove_unbalanced=remove_unbalanced,
            remove_changed=remove_changed,
            max_num_constraints=max_num_constraints,
            quiet=quiet,
        )
        self.open_elem = Element(open_elem)
        self.open_phases = [Composition(str(self.open_elem)).reduced_formula]
        self.mu = mu
        self.chempots = {self.open_elem: self.mu}
        self._build_grand_pd = True

    @staticmethod
    def _react_function(
        reactants, products, filtered_entries=None, pd=None, grand_pd=None, **kwargs
    ):
        """Same as the MinimizeGibbsEnumerator react function, but with ability to
        specify open element and grand potential phase diagram"""

        _ = products, kwargs  # unused

        r = list(reactants)
        r0 = r[0]

        if len(r) == 1:
            r1 = r[0]
        else:
            r1 = r[1]

        open_elem = list(grand_pd.chempots.keys())[0]

        for reactant in r:
            elems = reactant.composition.elements
            if len(elems) == 1 and elems[0] == open_elem:  # skip if reactant = open_e
                return []

        return react_interface(
            r0.composition,
            r1.composition,
            filtered_entries,
            pd,
            grand_pd=grand_pd,
        )



"""
Implements a class for storing a balanced reaction pathway.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from rxn_network.core import Composition
from rxn_network.pathways.base import Pathway
from rxn_network.pathways.basic import BasicPathway
from rxn_network.utils.funcs import limited_powerset

if TYPE_CHECKING:
    from rxn_network.reactions.base import Reaction


class BalancedPathway(BasicPathway):
    """
    Helper class for storing multiple ComputedReaction objects which form a single
    reaction pathway as identified via pathfinding methods. Includes costs for each
    reaction.
    """

    def __init__(
        self,
        reactions: list[Reaction],
        coefficients: list[float],
        costs: list[float],
        balanced: bool = False,
    ):
        """
        Args:
            reactions: list of ComputedReaction objects which occur along path.
            coefficients: list of coefficients to balance each corresponding reaction.
            costs: list of corresponding costs for each reaction.
            balanced: whether or not the reaction pathway is balanced.
                Defaults to False and should ideally be set through PathwaySolver.
        """
        self.coefficients = coefficients
        super().__init__(reactions=reactions, costs=costs)

        self.balanced = balanced

    def get_comp_matrix(self) -> np.ndarray:
        """
        Gets the composition matrix used in the balancing procedure.

        Returns:
            An array representing the composition matrix for a reaction
        """
        return np.array(
            [
                [
                    rxn.get_coeff(comp) if comp in rxn.all_comp else 0
                    for comp in self.compositions
                ]
                for rxn in self.reactions
            ]
        )

    def get_coeff_vector_for_rxn(self, rxn: Reaction) -> np.ndarray:
        """
        Gets the net reaction coefficients vector.

        Args:
            rxn: Reaction object to get coefficients for

        Returns:
            An array representing the reaction coefficients vector
        """
        return np.array(
            [
                rxn.get_coeff(comp) if comp in rxn.compositions else 0
                for comp in self.compositions
            ]
        )

    def contains_interdependent_rxns(self, precursors: list[Composition]) -> bool:
        """
        Whether or not the pathway contains interdependent reactions given a list of
        provided precursors.

        Args:
            precursors: List of precursor compositions
        """
        precursors_set = set(precursors)
        interdependent = False

        rxns = set(self.reactions)
        num_rxns = len(rxns)

        if num_rxns == 1:
            return False

        for combo in limited_powerset(rxns, num_rxns):
            size = len(combo)
            if (
                any(set(rxn.reactants).issubset(precursors_set) for rxn in combo)
                or size == 1
            ):
                continue

            other_comp = {c for rxn in (rxns - set(combo)) for c in rxn.compositions}

            unique_reactants = []
            unique_products = []
            for rxn in combo:
                unique_reactants.append(set(rxn.reactants) - precursors_set)
                unique_products.append(set(rxn.products) - precursors_set)

            overlap = [False] * size
            for i in range(size):
                for j in range(size):
                    if i == j:
                        continue
                    overlapping_phases = unique_reactants[i] & unique_products[j]
                    if overlapping_phases and (overlapping_phases not in other_comp):
                        overlap[i] = True

            if all(overlap):
                interdependent = True

        return interdependent

    @classmethod
    def balance(
        cls,
        pathway_sets: list[Pathway] | list[list[Reaction]],
        net_reaction: Reaction,
        tol: float = 1e-6,
    ):
        """Not implemented. See PathwaySolver class."""

        _, _, _ = pathway_sets, net_reaction, tol

        raise NotImplementedError(
            "Currently, to automatically balance and create a BalancedPathway object,"
            " you must use the PathwaySolver class."
        )

    @property
    def average_cost(self) -> float:
        """Returns the mean cost of the pathway"""
        return np.dot(self.coefficients, self.costs) / sum(self.coefficients)

    def __eq__(self, other) -> bool:
        if super().__eq__(other):
            return np.allclose(self.costs, other.costs)

        return False

    def __hash__(self):
        return hash((tuple(self.reactions), tuple(self.coefficients)))

    def __repr__(self) -> str:
        path_info = ""
        for rxn in self.reactions:
            path_info += f"{rxn} (dG = {round(rxn.energy_per_atom, 3)} eV/atom) \n"

        path_info += f"Average Cost: {round(self.average_cost,3)}"

        return path_info



"""
Implements an Entry that looks up pre-tabulated Gibbs free energies from the NIST-JANAF
tables.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from rxn_network.data import PATH_TO_FREED, load_experimental_data
from rxn_network.entries.experimental import ExperimentalReferenceEntry

if TYPE_CHECKING:
    from pymatgen.entries.computed_entries import EnergyAdjustment

    from rxn_network.core import Composition

G_COMPOUNDS = load_experimental_data(PATH_TO_FREED / "compounds.json.gz")


class FREEDReferenceEntry(ExperimentalReferenceEntry):
    """
    An Entry class for FREED experimental reference data. Given a composition,
    automatically finds the Gibbs free energy of formation, dGf(T) from tabulated
    reference values.

    Reference:
        https://www.thermart.net/freed-thermodynamic-database/
    """

    REFERENCES = G_COMPOUNDS

    def __init__(
        self,
        composition: Composition,
        temperature: float,
        energy_adjustments: list[EnergyAdjustment] | None = None,
        data: dict | None = None,
    ):
        """
        Args:
            composition: Composition object
            temperature: Temperature in Kelvin. If temperature is not selected from
                one of [300, 400, 500, ... 2000 K], then free energies will be
                interpolated. Defaults to 300 K.
            energy_adjustments: A list of EnergyAdjustments to apply to the entry.
            data: Optional dictionary containing entry data
        """
        super().__init__(
            composition=composition,
            temperature=temperature,
            energy_adjustments=energy_adjustments,
            data=data,
        )



"""Code for analyzing sets of reactions between two phases."""
from __future__ import annotations

from functools import cached_property, lru_cache
from itertools import combinations

import numpy as np
import plotly.express as px
from monty.json import MSONable
from plotly.graph_objs import Figure
from scipy.spatial import ConvexHull

from rxn_network.core import Composition
from rxn_network.reactions.computed import ComputedReaction


class InterfaceReactionHull(MSONable):
    """
    A class for storing and analyzing a set of reactions at an interface between two
    reactants. This class is more generalized than the InterfacialReactivity class and
    can encompass any set of reactions between two reactants, regardless of whether
    the reaction products are "stable" (i.e. together on the convex hull)
    """

    def __init__(
        self,
        c1: Composition,
        c2: Composition,
        reactions: list[ComputedReaction],
    ):
        """
        Args:
            c1: Composition of reactant 1
            c2: Composition of reactant 2
            reactions: List of reactions containing all enumerated reactions between the
                two reactants. Note that this list should not include identity reactions
                of the precursors.
        """
        self.c1 = Composition(c1).reduced_composition
        self.c2 = Composition(c2).reduced_composition
        self.e1 = None
        self.e2 = None

        for rxn in reactions:
            for e in rxn.reactant_entries:
                if e.composition.reduced_composition == self.c1:
                    self.e1 = e
                elif e.composition.reduced_composition == self.c2:
                    self.e2 = e
            if self.e1 is not None and self.e2 is not None:
                break
        else:
            raise ValueError(
                "Provided reactions do not correspond to reactant compositons!",
                c1,
                c2,
                reactions,
            )

        endpoint_reactions = [
            ComputedReaction.balance([self.e1], [self.e1]),
            ComputedReaction.balance([self.e2], [self.e2]),
        ]

        reactions_with_endpoints = reactions + endpoint_reactions

        coords = np.array(
            [(self.get_coordinate(r), r.energy_per_atom) for r in reactions]
        )
        coords = np.append(coords, [[0, 0], [1, 0]], axis=0)

        idx_sort = coords[:, 0].argsort()

        self.coords = coords[idx_sort]
        self.reactions = [reactions_with_endpoints[i] for i in idx_sort]
        self.hull = ConvexHull(self.coords)
        self.endpoint_reactions = endpoint_reactions

    def plot(self, y_max: float = 0.2) -> Figure:
        """
        Plot the reaction hull.
        """
        pts = self._get_scatter()
        lines = self._get_lines()

        fig = Figure(data=lines + [pts])

        fig.update_traces(
            hovertemplate=(
                "<b>%{hovertext}</b><br> <br><b>Atomic fraction</b>:"
                " %{x:.3f}<br><b>Energy</b>: %{y:.3f} (eV/atom)"
            )
        )
        fig.update_layout(yaxis_range=[min(self.coords[:, 1]) - 0.01, y_max])
        fig.update_layout(xaxis_title="Mixing ratio")
        fig.update_layout(yaxis_title="Energy (eV/atom)")

        return fig

    def get_energy_above_hull(self, reaction: ComputedReaction) -> float:
        """
        Get the energy of a reaction above the reaction hull.
        """
        idx = self.reactions.index(reaction)
        x, y = self.coords[idx]
        e_above_hull = y - self.get_hull_energy(x)

        return e_above_hull

    def get_coordinate(self, reaction: ComputedReaction) -> float:
        """Get coordinate of reaction in reaction hull. This is expressed as the atomic
        mixing ratio of component 2 in the reaction."""
        amt_c1 = reaction.reactant_atomic_fractions.get(self.c1, 0)
        amt_c2 = reaction.reactant_atomic_fractions.get(self.c2, 0)
        total = amt_c1 + amt_c2  # will add to 1.0 with two-component reactions
        try:
            coordinate = amt_c2 / total
        except ZeroDivisionError as e:
            raise ValueError(
                f"Can't compute coordinate for {reaction} with {self.c1}, {self.c2}"
            ) from e

        return round(coordinate, 12)  # avoids numerical issues

    def get_hull_energy(self, coordinate: float) -> float:
        """
        Get the energy of the reaction at a given coordinate.

        Args:
            coordinate: Coordinate of reaction in reaction hull.

        Returns:
            Energy of reaction at given coordinate.
        """
        reactions = self.get_reactions_by_coordinate(coordinate)
        return sum(weight * r.energy_per_atom for r, weight in reactions.items())

    def get_reactions_by_coordinate(
        self, coordinate: float
    ) -> dict[ComputedReaction, float]:
        """Get the reaction(s) at a given coordinate."""
        sorted_vertices = np.sort(self.hull_vertices)
        for i in range(len(sorted_vertices) - 1):
            v1 = sorted_vertices[i]
            v2 = sorted_vertices[i + 1]

            x1 = self.coords[v1, 0]
            x2 = self.coords[v2, 0]

            if np.isclose(coordinate, x1):
                return {self.reactions[v1]: 1.0}
            if np.isclose(coordinate, x2):
                return {self.reactions[v2]: 1.0}
            if x1 < coordinate < x2:
                return {
                    self.reactions[v1]: (x2 - coordinate) / (x2 - x1),
                    self.reactions[v2]: (coordinate - x1) / (x2 - x1),
                }

            continue

        raise ValueError("No reactions found!")

    def get_primary_competition(self, reaction: ComputedReaction) -> float:
        """
        Calculates the primary competition, C_1, for a reaction (in eV/atom).

        If you use this selectivity metric in your work, please cite the following work:

            McDermott, M. J.; McBride, B. C.; Regier, C.; Tran, G. T.; Chen, Y.; Corrao,
            A. A.; Gallant, M. C.; Kamm, G. E.; Bartel, C. J.; Chapman, K. W.; Khalifah,
            P. G.; Ceder, G.; Neilson, J. R.; Persson, K. A. Assessing Thermodynamic
            Selectivity of Solid-State Reactions for the Predictive Synthesis of
            Inorganic Materials. arXiv August 22, 2023.
            https://doi.org/10.48550/arXiv.2308.11816.

        Args:
            reaction: A computed reaction.

        Returns:
            The primary competition (C1 score) in eV/atom.
        """
        energy = reaction.energy_per_atom
        coord = self.get_coordinate(reaction)
        matching = np.where(self.coords[:, 0] == coord)[
            0
        ]  # remove all reactions at same coordinate

        idx_min = matching.min()
        idx_max = matching.max()

        competing_rxns = (
            self.reactions[:idx_min] + self.reactions[idx_max + 1 :]
        )  # assumes ordered

        competing_rxns = self.reactions[:idx_min] + self.reactions[idx_max + 1 :]
        competing_rxn_energies = [r.energy_per_atom for r in competing_rxns]
        min_energy = min(competing_rxn_energies)

        return energy - min_energy

    def get_secondary_competition(
        self,
        reaction: ComputedReaction,
        normalize: bool = True,
        include_e_hull: bool = False,
        recursive: bool = False,
    ) -> float:
        """
        Calculates the secondary competition, C_2, for a reaction (in eV/atom).

        If you use this selectivity metric in your work, please cite the following work:

            McDermott, M. J.; McBride, B. C.; Regier, C.; Tran, G. T.; Chen, Y.; Corrao,
            A. A.; Gallant, M. C.; Kamm, G. E.; Bartel, C. J.; Chapman, K. W.; Khalifah,
            P. G.; Ceder, G.; Neilson, J. R.; Persson, K. A. Assessing Thermodynamic
            Selectivity of Solid-State Reactions for the Predictive Synthesis of
            Inorganic Materials. arXiv August 22, 2023.
            https://doi.org/10.48550/arXiv.2308.11816.

        Args:
            reaction: A computed reaction.
            normalize: Whether or not to normalize the sum of secondary reaction
                sequence energies by the total number of sequnces. Defaults to True
                according to original definition.
            include_e_hull: Whether or not to include the energy above hull of the
                target reaction in the definition of C_2. According to the original
                paper, this defaults to False.
            recursive: Whether or not to perform the secondary reaction analysis via a
                recursive approach. This should return the same answer, but much slower.
                This argument exists for debugging / validation and defaults to False.

        Returns:
            The secondary competition (C2 score) in eV/atom.
        """
        x = self.get_coordinate(reaction)

        if not recursive:
            left_energy = self.get_decomposition_energy(0, x)
            left_num_paths = self.count(len(self.get_coords_in_range(0, x)) - 2)
            right_energy = self.get_decomposition_energy(x, 1)
            right_num_paths = self.count(len(self.get_coords_in_range(x, 1)) - 2)
        else:
            (
                left_energy,
                left_num_paths,
            ) = self.get_decomposition_energy_and_num_paths_recursive(0, x)
            (
                right_energy,
                right_num_paths,
            ) = self.get_decomposition_energy_and_num_paths_recursive(x, 1)

        if left_num_paths == 0:
            left_num_paths = 1
        if right_num_paths == 0:
            right_num_paths = 1

        energy = left_energy * right_num_paths + right_energy * left_num_paths
        total = left_num_paths * right_num_paths

        if normalize:
            energy = energy / total

        if include_e_hull:
            e_above_hull = self.get_energy_above_hull(reaction)
            energy -= e_above_hull

        return -1 * energy

    def get_secondary_competition_max_energy(self, reaction: ComputedReaction) -> float:
        """
        Calculates the score for a given reaction. This formula is based on a
        methodology presented in the following paper:

        Args:
            reaction: Reaction to calculate the competition score for.

        Returns:
            The competition score for the reaction
        """
        x = self.get_coordinate(reaction)
        left_energy = self.get_max_decomposition_energy(0, x)
        right_energy = self.get_max_decomposition_energy(x, 1)

        return -1 * (left_energy + right_energy)

    def get_secondary_competition_area(self, reaction: ComputedReaction) -> float:
        """
        Calculates the score for a given reaction. This formula is based on a
        methodology presented in the following paper: (TBD)

        Args:
            reaction: Reaction to calculate the competition score for.

        Returns:
            The competition score for the reaction
        """
        x = self.get_coordinate(reaction)
        left_area = self.get_decomposition_area(0, x)
        right_area = self.get_decomposition_area(x, 1)

        return left_area + right_area

    def get_max_decomposition_energy(self, x1: float, x2: float) -> float:
        coords = self.get_coords_in_range(x1, x2)

        max_energy = 0

        for c in combinations(range(len(coords)), 3):
            i_left, i_mid, i_right = sorted(c)
            c_left, c_mid, c_right = coords[[i_left, i_mid, i_right]]

            energy = self._calculate_altitude(c_left, c_mid, c_right)
            if energy < max_energy:
                max_energy = energy

        return max_energy

    def get_decomposition_energy(self, x1: float, x2: float) -> float:
        """
        Calculates the energy of the reaction decomposition between two points.

        Args:
            x1: Coordinate of first point.
            x2: Coordinate of second point.

        Returns:
            The energy of the reaction decomposition between the two points.
        """
        coords = self.get_coords_in_range(x1, x2)
        n = len(coords) - 2

        energy = 0
        for c in combinations(range(len(coords)), 3):
            i_left, i_mid, i_right = sorted(c)

            c_left, c_mid, c_right = coords[[i_left, i_mid, i_right]]

            n_left = (i_mid - i_left) - 1
            n_right = (i_right - i_mid) - 1

            count = self._altitude_multiplicity(n_left, n_right, n)
            energy += count * self._calculate_altitude(c_left, c_mid, c_right)

        return energy

    def get_decomposition_area(self, x1: float, x2: float) -> float:
        coords = self.get_coords_in_range(x1, x2)
        if len(coords) == 2:
            return 0

        return ConvexHull(
            coords, qhull_options="QJ i"
        ).volume  # this is how area is defined in scipy

    def get_coords_in_range(self, x1: float, x2: float):
        """
        Get the coordinates in the range [x1, x2].

        Args:
            x1: Start of range.
            x2: End of range.

        Returns:
            Array of coordinates in the range.

        """
        x_min, x_max = sorted([x1, x2])
        y_min, y_max = self.get_hull_energy(x_min), self.get_hull_energy(x_max)

        coords = []

        if np.isclose(x_min, 0.0) and not np.isclose(y_min, 0.0):
            coords.append([0.0, 0.0])

        coords.append([x_min, y_min])

        coords.extend(
            [
                self.coords[i]
                for i in self.hull_vertices
                if self.coords[i, 0] < x_max
                and self.coords[i, 0] > x_min
                and self.coords[i, 1] <= 0
            ]
        )

        if x_max != x_min:
            coords.append([x_max, y_max])
        if np.isclose(x_max, 1.0) and not np.isclose(y_max, 0.0):
            coords.append([1.0, 0.0])

        coords = np.array(coords)

        return coords[coords[:, 0].argsort()]  # type: ignore

    def count(self, num: int) -> int:
        """
        Reurns the number of decomposition pathways for the interface reaction hull
        based on the number of **product** vertices (i.e., total # of vertices
        considered - 2 reactant vertices).

        Precomputed for hulls up to size 15. Otherwise, calls recursive implementation
        of counting function.

        Args:
            num: Number of product vertices.
        """
        counts = [
            1,
            1,
            2,
            5,
            14,
            42,
            132,
            429,
            1430,
            4862,
            16796,
            58786,
            208012,
            742900,
            2674440,
            9694845,
            35357670,
            129644790,
            477638700,
            1767263190,
            6564120420,
            24466267020,
            91482563640,
        ]
        if num <= 22:
            count = counts[num]
        else:
            count = self._count_recursive(num)[0]

        return count

    @lru_cache(maxsize=128)
    def get_decomposition_energy_and_num_paths_recursive(
        self,
        x1: float,
        x2: float,
        use_x_min_ref: bool = True,
        use_x_max_ref: bool = True,
    ) -> tuple[float, int]:
        """
        This is a recursive implementation of the get_decomposition_energy function. It
        significantly slower than the non-recursive implementation but is more
        straightforward to understand. Both should return the same answer, however the
        refcursive implementation also includes "free" computation of the total number
        of paths. The function has been cached for speed.

        Args:
            x1: Coordinate of first point.
            x2: Coordinate of second point.
            use_x_min_ref: Useful for recursive calls. If true, uses the reactant at x=0
                as the reference (sometimes there is a decomposition reaction of the
                reactant that is lower in energy than the reactant).
            use_x_max_ref: Useful for recursive calls. If true, uses the reactant at
                x=1.0 as the reference (sometimes there is a decomposition reaction of
                the reactant that is lower in energy than the reactant).

        Returns:
            Tuple of decomposition energy and the number of decomposition pathways.
        """
        all_coords = self.get_coords_in_range(x1, x2)

        if not use_x_min_ref and all_coords[1, 0] == 0.0:
            all_coords = all_coords[1:]
        if not use_x_max_ref and all_coords[-1, 0] == 1.0:
            all_coords = all_coords[:-1]

        x_min, y_min = all_coords[0]
        x_max, y_max = all_coords[-1]

        coords = all_coords[1:-1, :]

        if len(coords) == 0:
            val, total = 0, 1
            return val, total
        if len(coords) == 1:
            val = self._calculate_altitude([x_min, y_min], coords[0], [x_max, y_max])
            total = 1
            return val, total

        val = 0
        total = 0
        for c in coords:
            if c[0] == 0.0 and not np.isclose(c[1], 0.0):
                use_x_min_ref = False
            elif c[0] == 1.0 and not np.isclose(c[1], 0.0):
                use_x_max_ref = False

            height = self._calculate_altitude([x_min, y_min], c, [x_max, y_max])
            (
                left_decomp,
                left_total,
            ) = self.get_decomposition_energy_and_num_paths_recursive(
                x_min, c[0], use_x_min_ref, use_x_max_ref
            )
            (
                right_decomp,
                right_total,
            ) = self.get_decomposition_energy_and_num_paths_recursive(
                c[0], x_max, use_x_min_ref, use_x_max_ref
            )

            val += (
                height * (left_total * right_total)
                + left_decomp * right_total
                + right_decomp * left_total
            )
            total += left_total * right_total

        return val, total

    @lru_cache(maxsize=128)
    def _altitude_multiplicity(self, n_left, n_right, n):
        """
        This function is used in the non-recursive implementation of the
        get_decomposition_energy function. It allows for rapid computation of the number
        times (multiplicitly) that a particular altitude appears in the decomposition.

        Args:
            n_left: number of vertices occurring in between the leftmost and middle
                vertices.
            n_right: number of vertices occurring in between the middle and
                rightmost vertices.
            n: total number of product vertices in the full interface reaction hull
                (i.e., not including the 2 reactant reference vertices )
        """
        remainder = n - n_left - n_right - 1
        if remainder < 0:
            return 0

        return self.count(n_left) * self.count(n_right) * self.count(remainder)

    def _count_recursive(self, n, cache=None):
        """
        A recursive implementation of the counting function.

        This implementation is courtesy of @mcgalcode.
        """
        if cache is None:
            cache = []
        if n == 0:
            return 1, [1]
        if n == 1:
            return 1, [1, 1]
        if len(cache) >= n:
            return cache[n], cache

        total = 0
        biggest_cache = []
        for i in range(n):
            left = i
            right = n - i - 1
            left_divs, c1 = self._count_recursive(left, biggest_cache)

            right_divs, c2 = self._count_recursive(right, biggest_cache)

            if len(c1) > len(biggest_cache):
                biggest_cache = c1

            if len(c2) > len(biggest_cache):
                biggest_cache = c2

            total += left_divs * right_divs
        return total, biggest_cache + [total]

    def _get_scatter(self):
        marker_size = 10

        pts = px.scatter(
            x=self.coords[:, 0],
            y=self.coords[:, 1],
            hover_name=[str(r) for r in self.reactions],
            labels={
                "x": "Mixing Ratio",
                "y": (
                    r"$\Delta G_{\mathrm{rxn}} ~"
                    r" \mathrm{\left(\dfrac{\mathsf{eV}}{\mathsf{atom}}\right)}$"
                ),
            },
        )
        pts.update_traces(marker={"size": marker_size})

        return pts.data[0]

    def _get_lines(self):
        coords = self.coords[self.hull.simplices]

        coords = coords[(coords[:, :, 1] <= 0).all(axis=1)]
        coords = coords[~(coords[:, :, 1] == 0).all(axis=1)]

        lines = [
            px.line(x=c[:, 0], y=c[:, 1]) for c in coords if not (c[:, 1] == 0).all()
        ]

        line_data = []
        for line in lines:
            line.update_traces(line={"color": "black"})
            line_data.append(line.data[0])

        return line_data

    @cached_property
    def hull_vertices(self) -> np.ndarray:
        hull_vertices = [
            i
            for i in self.hull.vertices
            if self.coords[i, 1] <= 0
            and np.isclose(
                self.coords[self.coords[:, 0] == self.coords[i, 0]][:, 1].min(),
                self.coords[i, 1],  # make sure point is lower than others on same x
            )
        ]

        return np.array(hull_vertices)

    @cached_property
    def stable_reactions(self) -> list[ComputedReaction]:
        """
        Returns the reactions that are stable (on the convex hull) of the interface
        reaction hull.
        """
        return [r for i, r in enumerate(self.reactions) if i in self.hull_vertices]

    @cached_property
    def unstable_reactions(self):
        """
        Returns the reactions that are unstable (NOT on the convex hull) of the
        interface reaction hull.
        """
        return [r for i, r in enumerate(self.reactions) if i not in self.hull_vertices]

    @staticmethod
    def _calculate_altitude(c_left, c_mid, c_right):
        """
        Helper geometry method: calculates the altitude of a point on a line defined by
        three points.

        Args:
            x1: point 1
            x2: point 2
            x3: point 3

        Returns:
            The altitude of the point
        """
        x1, y1 = c_left
        x2, y2 = c_mid
        x3, y3 = c_right

        xd = (x2 - x1) / (x3 - x1)
        yd = y1 + xd * (y3 - y1)

        return y2 - yd



"""
Basic interface for a reaction pathway.
"""
from __future__ import annotations

from abc import ABCMeta
from typing import TYPE_CHECKING

from monty.json import MSONable

if TYPE_CHECKING:
    from pymatgen.entries.computed_entries import ComputedEntry

    from rxn_network.core import Composition
    from rxn_network.reactions.base import Reaction


class Pathway(MSONable, metaclass=ABCMeta):
    """
    Base definition for a reaction pathway.
    """

    _reactions: list[Reaction]

    @property
    def entries(self) -> set[ComputedEntry]:
        """Entry objects in this pathway"""
        return {entry for rxn in self._reactions for entry in rxn.entries}

    @property
    def all_reactants(self) -> set[Composition]:
        """Reactant compositions for all reactions in the pathway"""
        return {entry for rxn in self._reactions for entry in rxn.reactants}

    @property
    def all_products(self) -> set[Composition]:
        """Product compositions reaction in the pathway"""
        return {entry for rxn in self._reactions for entry in rxn.products}

    @property
    def compositions(self) -> list[Composition]:
        """All compositions in the reaction"""
        return list(self.all_reactants | self.all_products)

    @property
    def reactants(self) -> set[Composition]:
        """The reactant compositions of this whole/net reaction pathway"""
        return self.all_reactants - self.all_products

    @property
    def products(self) -> set[Composition]:
        """The product compositions of this whole/net reaction pathway"""
        return self.all_products - self.all_reactants

    @property
    def intermediates(self) -> set[Composition]:
        """Intermediate compositions in this reaction pathway"""
        return self.all_products & self.all_reactants

    @property
    def energy(self) -> float:
        """Total energy of this reaction pathway"""
        return sum(rxn.energy for rxn in self._reactions)

    @property
    def energy_per_atom(self) -> float:
        """Total normalized energy of this reaction pathway"""
        return sum(rxn.energy_per_atom for rxn in self._reactions)



"""A calculator class for determining chemical potential distance of reactions."""
from __future__ import annotations

from itertools import chain, combinations, product
from typing import TYPE_CHECKING, Callable, Iterable

import numpy as np

from rxn_network.costs.base import Calculator
from rxn_network.thermo.chempot_diagram import ChemicalPotentialDiagram

if TYPE_CHECKING:
    from pymatgen.analysis.phase_diagram import PDEntry

    from rxn_network.reactions.computed import ComputedReaction
    from rxn_network.reactions.hull import InterfaceReactionHull


class ChempotDistanceCalculator(Calculator):
    """Calculator for determining the aggregated "chemical potential distance" for a
    reaction, in units of eV/atom. This is a reaction selectivity metric based on an
    aggregation function applied to the chemical potential differences of reactant-
    product and product-product interfaces in a reaction.

    If you use this cost metric, please cite the following work:

        Todd, P. K.; McDermott, M. J.; Rom, C. L.; Corrao, A. A.; Denney, J. J.;
        Dwaraknath, S. S.; Khalifah, P. G.; Persson, K. A.;  Neilson, J. R. Selectivity
        in Yttrium Manganese Oxide Synthesis via Local Chemical Potentials in
        Hyperdimensional Phase Space. J. Am. Chem. Soc. 2021, 143 (37), 15185-15194.
        https://doi.org/10.1021/jacs.1c06229.
    """

    def __init__(
        self,
        cpd: ChemicalPotentialDiagram,
        mu_func: Callable | str = "sum",
        name: str = "chempot_distance",
    ):
        """
        Args:
            cpd: the chemical potential diagram for the phase space in which the
                reaction(s) exist
            mu_func: the function (or string name of the function) used to aggregate the
                interfacial chemical potential distances into a single value describing
                the whole reaction. Current options are 1) max, 2) mean, and 3) sum
                (default).
            name: the data dictionary key with which to store the calculated value.
                Defaults to "chempot_distance".
        """
        self.cpd = cpd
        self.name = name

        if mu_func == "max":
            self.mu_func = max
        elif mu_func == "mean":
            self.mu_func = np.mean  # type: ignore
        elif mu_func == "sum":
            self.mu_func = sum  # type: ignore
        elif isinstance(mu_func, str):
            raise ValueError(
                "Provided mu_func name is not a known function; please provide the"
                " function directly."
            )

        self._open_elems = set()
        if cpd.entries[0].__class__.__name__ == "GrandPotPDEntry":
            self._open_elems = set(cpd.entries[0].chempots.keys())

    def calculate(self, rxn: ComputedReaction) -> float:
        """Calculates the aggregate chemical potential distance in eV/atom. The mu_func
        parameter determines how the individual pairwise interface distances are
        aggregated into a single value describing the overall reaction. When mu_func =
        "sum" (i.e., the default setting), the total chemical potential distance is
        returned.

        Args:
            rxn: the ComputedReaction object

        Returns:
            The aggregate chemical potential distance of the reaction in eV/atom.
        """
        reactant_entries = rxn.reactant_entries
        product_entries = rxn.product_entries

        if hasattr(rxn, "grand_entries"):
            reactant_entries = [
                e
                for e, c in zip(rxn.grand_entries, rxn.coefficients)
                if c < 0 and e.__class__.__name__ == "GrandPotPDEntry"
            ]
            product_entries = [
                e
                for e, c in zip(rxn.grand_entries, rxn.coefficients)
                if c > 0 and e.__class__.__name__ == "GrandPotPDEntry"
            ]
        combos = chain(
            product(reactant_entries, product_entries),
            combinations(product_entries, 2),
        )
        distances = [
            self.cpd.shortest_domain_distance(
                combo[0].composition.reduced_formula,
                combo[1].composition.reduced_formula,
                offset=self.cpd.get_offset(combo[0]) + self.cpd.get_offset(combo[1]),
            )
            for combo in combos
        ]

        distance = round(float(self.mu_func(distances)), 5)
        return distance

    @classmethod
    def from_entries(
        cls,
        entries: list[PDEntry],
        mu_func: Callable[[Iterable[float]], float] | str = "sum",
        name: str = "chempot_distance",
        **kwargs,
    ) -> ChempotDistanceCalculator:
        """Convenience constructor which first builds the ChemicalPotentialDiagram
        object from a list of entry objects.

        Args:
            entries: entry objects used to build the ChemicalPotentialDiagram
            mu_func: the name of the function used to process the interfacial
                chemical potential distances into a single value describing the whole
                reaction.
            name: the data dictionary key by which to store the calculated value,
                defaults to "chempot_distance"
            **kwargs: optional kwargs passed to ChemicalPotentialDiagram.

        Returns:
            A ChempotDistanceCalculator object
        """

        cpd = ChemicalPotentialDiagram(entries=entries, **kwargs)
        return cls(cpd, mu_func, name)


class PrimaryCompetitionCalculator(Calculator):
    """Calculator for determining the primary competition, C_1, for a reaction (units:
    eV/atom).

    If you use this selectivity metric in your work, please cite the following work:

        McDermott, M. J.; McBride, B. C.; Regier, C.; Tran, G. T.; Chen, Y.; Corrao, A.
        A.; Gallant, M. C.; Kamm, G. E.; Bartel, C. J.; Chapman, K. W.; Khalifah, P. G.;
        Ceder, G.; Neilson, J. R.; Persson, K. A. Assessing Thermodynamic Selectivity of
        Solid-State Reactions for the Predictive Synthesis of Inorganic Materials. arXiv
        August 22, 2023. https://doi.org/10.48550/arXiv.2308.11816.
    """

    def __init__(
        self,
        irh: InterfaceReactionHull,
        name: str = "primary_competition",
    ):
        """
        Args:
            irh: the interface reaction hull containing the target reaction and all
                competing reactions.
            name: the data dictionary key with which to store the calculated value.
                Defaults to "primary_competition".
        """
        self.irh = irh
        self.name = name

    def calculate(self, rxn: ComputedReaction) -> float:
        """Calculates the competitiveness score for a given reaction by enumerating
        competing reactions, evaluating their cost with the supplied cost function, and
        then using the c-score formula, i.e. the _get_c_score() method, to determine the
        competitiveness score.

        Args:
            rxn: the ComputedReaction object to be evaluated

        Returns:
            The C1 score
        """
        return self.irh.get_primary_competition(rxn)


class SecondaryCompetitionCalculator(Calculator):
    """Calculator for determining the secondary competition, C_2, for a reaction (in
    eV/atom).

    If you use this selectivity metric in your work, please cite the following work:

        McDermott, M. J.; McBride, B. C.; Regier, C.; Tran, G. T.; Chen, Y.; Corrao, A.
        A.; Gallant, M. C.; Kamm, G. E.; Bartel, C. J.; Chapman, K. W.; Khalifah, P. G.;
        Ceder, G.; Neilson, J. R.; Persson, K. A. Assessing Thermodynamic Selectivity of
        Solid-State Reactions for the Predictive Synthesis of Inorganic Materials. arXiv
        August 22, 2023. https://doi.org/10.48550/arXiv.2308.11816.
    """

    def __init__(
        self,
        irh: InterfaceReactionHull,
        name: str = "secondary_competition",
    ):
        """
        Args:
            irh: the interface reaction hull containing the target reaction and all
                competing reactions.
            name: the data dictionary key with which to store the calculated value.
                Defaults to "secondary_competition".
        """
        self.irh = irh
        self.name = name

    def calculate(self, rxn: ComputedReaction) -> float:
        """Calculates the secondary competition per its implementation in the
        InterfaceReactionHull class.

        Args:
            rxn: the ComputedReaction object to be evaluated

        Returns:
            The C2 score
        """
        return self.irh.get_secondary_competition(rxn)


class SecondaryCompetitionWithEhullCalculator(Calculator):
    """
    WARNING: this is an alternative calculator for secondary competition (C_2) that
    includes the energy above hull of the target reaciton. It should only be used for
    testing purposes.
    """

    def __init__(
        self,
        irh: InterfaceReactionHull,
        name: str = "secondary_competition_with_ehull",
    ):
        """
        Args:
            irh: the interface reaction hull containing the target reaction and all
                competing reactions.
            name: the data dictionary key with which to store the calculated value.
                Defaults to "secondary_competition_with_ehull".
        """
        self.irh = irh
        self.name = name

    def calculate(self, rxn: ComputedReaction) -> float:
        """Calculates the secondary competition with e_hull per its implementation in
        the InterfaceReactionHull class.

        Args:
            rxn: the ComputedReaction object to be evaluated

        Returns:
            The C2 + e_hull score
        """
        return self.irh.get_secondary_competition(rxn, include_e_hull=True)


class SecondaryCompetitionMaxCalculator(Calculator):
    """
    WARNING: this is an alternative calculator for secondary competition (C_2) that
    defaults to calculation of the maximum secondary reaction energy. It should only be
    used for testing purposes.
    """

    def __init__(
        self,
        irh: InterfaceReactionHull,
        name: str = "secondary_competition_max",
    ):
        """
        Args:
            irh: the interface reaction hull containing the target reaction and all
                competing reactions.
            name: the data dictionary key with which to store the calculated value.
                Defaults to "secondary_competition_max".
        """
        self.irh = irh
        self.name = name

    def calculate(self, rxn: ComputedReaction) -> float:
        """Calculates the secondary competition with max energies.

        Args:
            rxn: the ComputedReaction object to be evaluated

        Returns:
            The C2 score using max energies.
        """
        return self.irh.get_secondary_competition_max_energy(rxn)


class SecondaryCompetitionAreaCalculator(Calculator):
    """
    WARNING: this is an alternative calculator for secondary competition (C_2) that
    defaults to calculation of the area of the enclosed hull. It should only be used for
    testing purposes and is quite unstable.
    """

    def __init__(
        self,
        irh: InterfaceReactionHull,
        name: str = "secondary_competition_area",
    ):
        """
        Args:
            irh: the interface reaction hull containing the target reaction and all
                competing reactions.
            name: the data dictionary key with which to store the calculated value.
                Defaults to "secondary_competition_area".
        """
        self.irh = irh
        self.name = name

    def calculate(self, rxn: ComputedReaction) -> float:
        """Calculates an area for secondary competition.

        Args:
            rxn: the ComputedReaction object to be evaluated

        Returns:
            Secondary competition as represented by area.
        """
        return self.irh.get_secondary_competition_area(rxn)



"""Composition class used to represent a chemical composition."""
from functools import cached_property

from pymatgen.core.composition import Composition as PymatgenComposition


class Composition(PymatgenComposition):
    """
    Modified Composition class adapted from pymatgen.

    The purpose of this is to modify / extend methods for better performance within the
    rxn_network package.

    """

    def __init__(
        self, *args, strict: bool = False, **kwargs
    ):  # pylint: disable=useless-parent-delegation
        super().__init__(*args, strict=strict, **kwargs)

    @cached_property
    def reduced_formula(self) -> str:
        """
        Returns a pretty normalized formula, i.e., LiFePO4 instead of
        Li4Fe4P4O16.
        """
        return self.get_reduced_formula_and_factor()[0]



"""Class for intepolated entries"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pymatgen.analysis.phase_diagram import GrandPotPDEntry
from pymatgen.entries.computed_entries import ComputedEntry

from rxn_network.core import Composition

if TYPE_CHECKING:
    from pymatgen.core.periodic_table import Element


class InterpolatedEntry(ComputedEntry):
    """
    Lightweight Entry object for computed data. Contains facilities
    for applying corrections to the energy attribute and for storing
    calculation parameters.
    """

    def __init__(
        self,
        composition: Composition,
        energy: float,
        correction: float = 0.0,
        energy_adjustments: list | None = None,
        parameters: dict | None = None,
        data: dict | None = None,
        entry_id: object | None = None,
    ):
        """

        Args:
            composition: Composition of the entry. For
                flexibility, this can take the form of all the typical input
                taken by a Composition, including a {symbol: amt} dict,
                a string formula, and others.
            energy: Energy of the entry. Usually the final calculated
                energy from VASP or other electronic structure codes.
            correction: Manually set an energy correction, will ignore
                energy_adjustments if specified. Defaults to 0.
            energy_adjustments: An optional list of EnergyAdjustment to
                be applied to the energy. This is used to modify the energy for
                certain analyses. Defaults to None.
            parameters: An optional dict of parameters associated with
                the entry. Defaults to None.
            data: An optional dict of any additional data associated
                with the entry. Defaults to None.
            entry_id: An optional id to uniquely identify the entry.
        """
        composition = Composition(composition)

        if entry_id is None:
            entry_id = f"{self.__class__.__name__}-{composition.formula}"

        super().__init__(
            composition,
            energy,
            correction=correction,
            energy_adjustments=energy_adjustments,
            parameters=parameters,
            data=data,
            entry_id=entry_id,
        )

    def to_grand_entry(self, chempots: dict[Element, float]) -> GrandPotPDEntry:
        """
        Convert a GibbsComputedEntry to a GrandComputedEntry.

        Args:
            chempots: A dictionary of {element: chempot} pairs.

        Returns:
            A GrandComputedEntry.
        """
        return GrandPotPDEntry(self, chempots)

    @property
    def unique_id(self) -> str:
        """
        Returns a unique ID for the entry.
        """
        return self.entry_id

    @property
    def is_experimental(self) -> bool:
        """Returns True by default."""
        return False

    def __repr__(self):
        output = [
            (
                f"InterpolatedEntry | {self.composition.formula} "
                f"({self.composition.reduced_formula})"
            ),
            f"Energy  = {self.energy:.4f}",
        ]
        return "\n".join(output)

    def __eq__(self, other) -> bool:
        if not type(other) is type(self):
            return False

        if not np.isclose(self.energy, other.energy):
            return False

        if getattr(self, "entry_id", None) and getattr(other, "entry_id", None):
            return self.entry_id == other.entry_id

        if self.composition != other.composition:
            return False

        return True

    def __hash__(self):
        return hash(
            (
                self.composition,
                self.energy,
                self.entry_id,
            )
        )



"""
A reaction class that builds reactions based on ComputedEntry objects and provides
information about reaction thermodynamics.
"""
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.core.composition import Element
from uncertainties import ufloat

from rxn_network.core import Composition
from rxn_network.reactions.basic import BasicReaction

if TYPE_CHECKING:
    from pymatgen.entries.computed_entries import ComputedEntry


class ComputedReaction(BasicReaction):
    """
    Convenience class to generate a reaction from ComputedEntry objects, with
    some additional attributes, such as a reaction energy based on computed
    energies. This class also balances the reaction.
    """

    def __init__(
        self,
        entries: list[ComputedEntry],
        coefficients: np.ndarray | list[float],
        data: dict | None = None,
        lowest_num_errors: int | float = 0,
    ):
        """
        Args:
            entries: List of ComputedEntry objects.
            coefficients: List or array of reaction coefficients.
            data: Optional dict of data
            lowest_num_errors: number of "errors" encountered during reaction balancing.
        """
        self._entries = list(entries)
        self.reactant_entries = [
            entry for entry, coeff in zip(entries, coefficients) if coeff < 0
        ]
        self.product_entries = [
            entry for entry, coeff in zip(entries, coefficients) if coeff > 0
        ]
        compositions = [e.composition.reduced_composition for e in entries]

        super().__init__(
            compositions, coefficients, data=data, lowest_num_errors=lowest_num_errors
        )

    @classmethod
    def balance(
        cls,
        reactant_entries: list[ComputedEntry],
        product_entries: list[ComputedEntry],
        data: dict | None = None,
    ) -> ComputedReaction:
        """
        Balances and returns a new ComputedReaction.

        Reactants and products to be specified as a collection (list, set, etc.) of
        ComputedEntry objects.

        Args:
            reactant_entries: List of reactant entries
            product_entries: List of product entries
            data: Optional dict of data
        """
        reactant_comps = [e.composition.reduced_composition for e in reactant_entries]
        product_comps = [e.composition.reduced_composition for e in product_entries]
        coefficients, lowest_num_errors, num_constraints = cls._balance_coeffs(
            reactant_comps, product_comps
        )

        if not data:
            data = {}
        data["num_constraints"] = num_constraints

        return cls(
            entries=list(reactant_entries) + list(product_entries),
            coefficients=coefficients,
            data=data,
            lowest_num_errors=lowest_num_errors,
        )

    def get_new_temperature(self, new_temperature: float) -> ComputedReaction:
        """
        Returns a new reaction with the temperature changed.

        Args:
            new_temperature: New temperature in Kelvin
        """
        try:
            new_entries = [e.get_new_temperature(new_temperature) for e in self.entries]
        except AttributeError as e:
            raise AttributeError(
                "One or more of the entries in the reaction is not associated with a"
                " temperature. Please use the GibbsComputedEntry class for all entries"
                " in the reaction."
            ) from e

        return ComputedReaction(
            new_entries,
            self.coefficients,
            data=self.data,
            lowest_num_errors=self.lowest_num_errors,
        )

    @cached_property
    def energy(self) -> float:
        """
        Returns (float):
            The calculated reaction energy.
        """
        calc_energies: dict[Composition, float] = {}

        for entry in self._entries:
            (comp, factor) = entry.composition.get_reduced_composition_and_factor()
            calc_energies[comp] = min(
                calc_energies.get(comp, float("inf")), entry.energy / factor
            )

        return sum(
            amt * calc_energies[c]
            for amt, c in zip(self.coefficients, self.compositions)
        )

    @cached_property
    def energy_per_atom(self) -> float:
        """
        Returns (float):
            The calculated reaction energy in eV, divided by the total number of atoms
            in the reaction. Cached for speedup.
        """
        return self.energy / self.num_atoms

    @cached_property
    def energy_uncertainty(self) -> float:
        """
        Calculates the uncertainty in the reaction energy based on the uncertainty in
        the energies of the reactants/products. Cached for speedup.
        """

        calc_energies: dict[Composition, ufloat] = {}

        for entry in self._entries:
            (comp, factor) = entry.composition.get_reduced_composition_and_factor()
            energy_ufloat = ufloat(entry.energy, entry.correction_uncertainty)
            calc_energies[comp] = min(
                calc_energies.get(comp, float("inf")), energy_ufloat / factor
            )

        energy_with_uncertainty = sum(
            amt * calc_energies[c]
            for amt, c in zip(self.coefficients, self.compositions)
        )

        return energy_with_uncertainty.s  # type: ignore

    @cached_property
    def energy_uncertainty_per_atom(self) -> float:
        """
        Returns the energy_uncertainty divided by the total number of atoms in the
        reaction.
        """
        return self.energy_uncertainty / self.num_atoms

    @property
    def entries(self) -> list[ComputedEntry]:
        """
        Returns a copy of the entries
        """
        return self._entries

    def copy(self) -> ComputedReaction:
        """
        Returns a copy of the Reaction object
        """
        return ComputedReaction(
            self.entries, self.coefficients, self.data, self.lowest_num_errors
        )

    def reverse(self) -> ComputedReaction:
        """
        Returns a reversed reaction (i.e. sides flipped)

        """
        return ComputedReaction(
            self.entries, -1 * self.coefficients, self.data, self.lowest_num_errors
        )

    def normalize_to(self, comp: Composition, factor: float = 1) -> ComputedReaction:
        """
        Normalizes the reaction to one of the compositions via the provided factor.

        By default, normalizes such that the composition given has a coefficient of
        1.

        Args:
            comp: Composition object to normalize to
            factor: factor to normalize to. Defaults to 1.
        """
        coeffs = self.coefficients.copy()
        scale_factor = abs(1 / coeffs[self.compositions.index(comp)] * factor)
        coeffs *= scale_factor
        return ComputedReaction(self.entries, coeffs, self.data, self.lowest_num_errors)

    def normalize_to_element(
        self, element: Element, factor: float = 1
    ) -> ComputedReaction:
        """
        Normalizes the reaction to one of the elements.
        By default, normalizes such that the amount of the element is 1.
        Another factor can be specified.

        Args:
            element (Element/Species): Element to normalize to.
            factor (float): Factor to normalize to. Defaults to 1.
        """
        all_comp = self.compositions
        coeffs = self.coefficients.copy()
        current_el_amount = (
            sum(all_comp[i][element] * abs(coeffs[i]) for i in range(len(all_comp))) / 2
        )
        scale_factor = factor / current_el_amount
        coeffs *= scale_factor
        return ComputedReaction(self.entries, coeffs, self.data, self.lowest_num_errors)

    def get_entry_idx_vector(self, n) -> np.ndarray:
        indices = [e.data["idx"] for e in self.entries]

        v = np.zeros(n)
        v[indices] = self.coefficients
        return v

    def __eq__(self, other) -> bool:
        is_equal = BasicReaction.__eq__(self, other)

        if is_equal:
            is_equal = np.isclose(  # type: ignore
                self.energy_per_atom, other.energy_per_atom
            )

        return is_equal

    def __hash__(self):
        """Must be redefined due to overriding __eq__"""
        return hash(
            (self.chemical_system, tuple(sorted(self.coefficients)))
        )  # not checking here for reactions that are multiples (too expensive)



"""
Implementation of cost functions used in the package (e.g., Softplus, WeightedSum).
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from rxn_network.costs.base import CostFunction

if TYPE_CHECKING:
    from rxn_network.reactions.computed import ComputedReaction


class Softplus(CostFunction):
    """
    The softplus cost function is a smoothed version of the Rectified Linear Unit (ReLU)
    function commonly used in neural networks. It has the property that the output goes
    to 0 as the input goes to negative infinity, but the output approaches a linear
    scaling as the input goes to positive infinity. This is an especially useful mapping
    for determining a cost ranking of a reaction.
    """

    def __init__(
        self,
        temp: float = 300,
        params: list[str] | None = None,
        weights: list[float] | None = None,
    ):
        """
        Args:
            temp: temperature in Kelvin [K]. This serves as a scale factor for the
                output of the function. Higher temperatures -> lower costs. Defaults to
                300 K.
            params: List of data dictionary keys for function parameters used as an
                argument to the softplus function. Defaults to ["energy_per_atom"]
            weights: List of corresponding values by which to weight the
                function parameters. Defaults to [1.0].
        """
        if params is None:
            params = ["energy_per_atom"]
        if weights is None:
            weights = [1.0]
        if temp <= 0:
            raise ValueError("Temperature must be greater than zero!")

        self.temp = temp
        self.params = params
        self.weights = np.array(weights)

    def evaluate(self, rxn: ComputedReaction) -> float:
        """
        Calculates the cost of reaction based on the initialized parameters and weights.

        Args:
            rxn: A ComputedReaction to evaluate.

        Returns:
            The cost of the reaction.
        """
        values = []
        for p in self.params:
            if rxn.data and p in rxn.data:
                value = rxn.data[p]
            elif hasattr(rxn, p):
                value = getattr(rxn, p)
            else:
                raise ValueError(f"Reaction is missing parameter {p}!")
            values.append(value)

        values_arr = np.array(values)
        total = float(np.dot(values_arr, self.weights))

        return self._softplus(total, self.temp)

    @staticmethod
    def _softplus(x: float, t: float) -> float:
        """The mathematical formula for the softplus function"""
        return np.log(1 + (273 / t) * np.exp(x))

    def __repr__(self):
        return (
            "Softplus with parameters: "
            f"{' '.join([f'{k} ({v})' for k, v in zip(self.params, self.weights)])}"
        )


class WeightedSum(CostFunction):
    """
    This a weighted sum of user-specified parameters.
        cost = param_1*weight_1 + param_2*weight_2 + param_3*weight_3 ...
    """

    def __init__(
        self,
        params: list[str] | None = None,
        weights: list[float] | None = None,
    ):
        """
        Args:
            params: List of data dictionary keys for function parameters used as an
                argument to the weighted summation. Defaults to ["energy_per_atom"]
            weights: List of corresponding values by which to weight the
                function parameters. Defaults to [1.0].
        """
        if params is None:
            params = ["energy_per_atom"]
        if weights is None:
            weights = [1.0]

        self.params = params
        self.weights = np.array(weights)

    def evaluate(self, rxn: ComputedReaction) -> float:
        """
        Calculates the cost of reaction based on the initialized parameters and weights.

        Args:
            rxn: A ComputedReaction to evaluate.

        Returns:
            The cost of the reaction.
        """
        values = []
        for p in self.params:
            if rxn.data and p in rxn.data:
                value = rxn.data[p]
            elif hasattr(rxn, p):
                value = getattr(rxn, p)
            else:
                raise ValueError(f"Reaction is missing parameter {p}!")
            values.append(value)

        values_arr = np.array(values)
        total = float(np.dot(values_arr, self.weights))

        return total

    def __repr__(self):
        return (
            "Weighted sum with parameters: "
            f"{' '.join([f'{k} ({v})' for k, v in zip(self.params, self.weights)])}"
        )



"""
Implements an Entry that looks up NIST pre-tabulated Gibbs free energies
"""
from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING

from monty.json import MontyDecoder
from pymatgen.analysis.phase_diagram import GrandPotPDEntry
from pymatgen.entries.computed_entries import ComputedEntry, EnergyAdjustment
from scipy.interpolate import interp1d

from rxn_network.core import Composition

if TYPE_CHECKING:
    from pymatgen.core.periodic_table import Element


class ExperimentalReferenceEntry(ComputedEntry):
    """
    An Entry class for experimental reference data, to be sub-classed for specific data
    sources.  Given a composition, automatically finds the Gibbs free energy of
    formation, dGf(T) from tabulated reference values.
    """

    REFERENCES: dict = {}

    def __init__(
        self,
        composition: Composition,
        temperature: float,
        energy_adjustments: list[EnergyAdjustment] | None = None,
        data: dict | None = None,
    ):
        """
        Args:
            composition: Composition object
            temperature: Temperature in Kelvin. If temperature is not selected within
                the range of the reference  data (see self._validate_temperature), then
                this will raise an error.
            energy_adjustments: A list of EnergyAdjustments to apply to the entry.
            data: Optional dictionary containing entry data
        """
        formula = composition.reduced_formula
        entry_id = f"{self.__class__.__name__}-{formula}_{temperature}"

        self._temperature = temperature
        self._validate_temperature(formula, temperature)

        energy = self._get_energy(formula, temperature)

        super().__init__(
            composition,
            energy,
            energy_adjustments=energy_adjustments,
            data=data,
            entry_id=entry_id,
        )
        self._composition = composition

        self.name = formula

    def get_new_temperature(self, new_temperature: float) -> ExperimentalReferenceEntry:
        """
        Return a copy of the NISTReferenceEntry at the new specified temperature.

        Args:
            new_temperature: The new temperature to use [K]

        Returns:
            A copy of the NISTReferenceEntry at the new specified temperature.
        """
        new_entry_dict = self.as_dict()
        new_entry_dict["temperature"] = new_temperature

        new_entry = self.from_dict(new_entry_dict)
        return new_entry

    def to_grand_entry(self, chempots: dict[Element, float]):
        """
        Convert an ExperimentalReferenceEntry to a GrandComputedEntry.

        Args:
            chempots: A dictionary of {element: chempot} pairs.

        Returns:
            A GrandComputedEntry.
        """
        return GrandPotPDEntry(self, chempots)

    @classmethod
    def _validate_temperature(cls, formula: str, temperature: float) -> None:
        """Ensure that the temperature is from a valid range."""
        if formula not in cls.REFERENCES:
            raise ValueError(f"{formula} not in reference data!")

        g = cls.REFERENCES[formula]

        if temperature < min(g) or temperature > max(g):
            raise ValueError(
                f"Temperature must be selected from range: {min(g)} K to {max(g)} K"
            )

    @classmethod
    def _get_energy(cls, formula: str, temperature: float) -> float:
        """
        Convenience method for accessing and interpolating experimental data.

        Args:
            formula: Chemical formula by which to search experimental data.
            temperature: Absolute temperature [K].

        Returns:
            Gibbs free energy of formation of formula at specified temperature [eV]
        """
        data = cls.REFERENCES[formula]

        if temperature % 100 > 0:
            g_interp = interp1d(list(data.keys()), list(data.values()))
            return g_interp(temperature)[()]

        return data[temperature]

    @property
    def temperature(self) -> float:
        """Returns temperature used to calculate entry's energy"""
        return self._temperature

    @property
    def is_experimental(self) -> bool:
        """Returns True by default."""
        return True

    @property
    def is_element(self) -> bool:
        """Returns True if the entry is an element."""
        return self.composition.is_element

    @property
    def unique_id(self) -> str:
        """
        Returns a unique ID for the entry based on its entry-id and temperature. This is
        useful because the same entry-id can be used for multiple entries at different
        temperatures.
        """
        return self.entry_id

    def as_dict(self) -> dict:
        """
        Returns:
            A dict representation of the Entry.
        """
        d = super().as_dict()

        d["temperature"] = self.temperature

        del d["energy"]
        del d["entry_id"]
        del d["parameters"]
        del d["correction"]

        return d

    @classmethod
    def from_dict(cls, d) -> ExperimentalReferenceEntry:
        dec = MontyDecoder()
        entry = cls(
            composition=Composition(d["composition"]),
            temperature=d["temperature"],
            energy_adjustments=dec.process_decoded(d["energy_adjustments"]),
            data=d["data"],
        )
        return entry

    def __repr__(self) -> str:
        output = [
            f"{self.__class__.__name__} | {self.composition.reduced_formula}",
            f"Gibbs Energy ({self.temperature} K) = {self.energy:.4f}",
        ]
        return "\n".join(output)

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return (
                (self.composition.reduced_formula == other.composition.reduced_formula)
                and (self.temperature == other.temperature)
                and (set(self.energy_adjustments) == set(other.energy_adjustments))
            )
        return False

    def __hash__(self) -> int:
        data_md5 = hashlib.md5(  # nosec
            f"{self.__class__.__name__}{self.composition}_{self.temperature}".encode(
                "utf-8"
            )
        ).hexdigest()
        return int(data_md5, 16)



"""Definitions of common job functions"""
from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from pymatgen.core.composition import Element

from rxn_network.core import Composition
from rxn_network.utils.funcs import get_logger

if TYPE_CHECKING:
    from rxn_network.entries.entry_set import GibbsEntrySet

logger = get_logger(__name__)


def get_added_elem_data(
    entries: GibbsEntrySet, targets: Iterable[Composition | str]
) -> tuple[list[Element], str]:
    """
    Given a provided entry set and targets, this identifies which elements in the entry
    set are "additional" (not found in the target)

    Args:
        entries: the full entry set
        targets: the target phase compositions

    Returns:
        A tuple of the additional elements and their chemical system string.
    """
    added_elems = entries.chemsys - {
        str(e) for target in targets for e in Composition(target).elements
    }
    added_chemsys = "-".join(sorted(list(added_elems)))
    added_elements = [Element(e) for e in added_elems]

    return added_elements, added_chemsys
